import glob
import os
import shlex
import tarfile
import tempfile
import zipfile
from typing import Iterable, List, Optional, Set, Union

from redun.executors.scratch import get_code_scratch_file
from redun.file import File
from redun.hashing import hash_stream


def find_code_files(
    basedir: str = ".",
    includes: Optional[List[str]] = None,
    excludes: Optional[List[str]] = None,
) -> Iterable[str]:
    """
    Find all the workflow code files consistent with the include/exclude patterns.
    """
    if includes is None:
        includes = ["**/*.py"]
    if excludes is None:
        excludes = []

    files: Set[str] = set()
    for pattern in includes:
        files |= set(glob.glob(os.path.join(basedir, pattern), recursive=True))

    for pattern in excludes:
        files -= set(glob.glob(os.path.join(basedir, pattern), recursive=True))
    return files


def create_tar(
    tar_path: str, file_paths: Iterable[str], arcname_prefix: Optional[str] = None
) -> File:
    """
    Create a tar file from local file paths.
    Args:
        arcname_prefix: prefix to add to each file path in the tar file.
    """
    tar_file = File(tar_path)

    with tar_file.open("wb") as out:
        with tarfile.open(fileobj=out, mode="w|gz") as tar:
            for file_path in file_paths:
                # add prefix to each file
                arcname = os.path.join(arcname_prefix, file_path) if arcname_prefix else file_path
                tar.add(file_path, arcname=arcname)

    return tar_file


def extract_tar(tar_file: File, dest_dir: str = ".") -> None:
    """
    Extract a tar file to local paths.
    """
    with tar_file.open("rb") as infile:
        with tarfile.open(fileobj=infile, mode="r|gz") as tar:
            tar.extractall(dest_dir)


def create_zip(
    zip_path: str,
    base_path: str,
    file_paths: Iterable[str],
    arcname_prefix: Optional[str] = None,
) -> File:
    """
    Create a zip file from local file paths.
    Args:
        arcname_prefix: prefix to add to each file path in the tar file.
    """
    zip_file = File(zip_path)

    with zip_file.open("wb") as out:
        with zipfile.ZipFile(out, mode="w") as stream:
            for file_path in file_paths:
                arcname = os.path.relpath(file_path, base_path)
                arcname = os.path.join(arcname_prefix, arcname) if arcname_prefix else arcname
                stream.write(file_path, arcname)

    return zip_file


def parse_code_package_config(config) -> Union[dict, bool]:
    """
    Parse the code package options from a AWSBatchExecutor config.
    """
    if not config.getboolean("code_package", fallback=True):
        return False

    include_config = config.get("code_includes", "**/*.py")
    exclude_config = config.get("code_excludes", "")

    return {
        "includes": shlex.split(include_config),
        "excludes": shlex.split(exclude_config),
    }


def package_code(
    scratch_prefix: str,
    code_package: dict = {},
    use_zip: bool = False,
    basename: Optional[str] = None,
    arcname_prefix: Optional[str] = None,
) -> File:
    """
    Package code to scratch directory.
    Args:
        basename: If provided, uses this string as the basename instead of the
            calculated tarball hash.
        arcname_prefix: Optional suffix to append to tarball basename.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        file_paths = find_code_files(
            includes=code_package.get("includes"), excludes=code_package.get("excludes")
        )

        if use_zip:
            temp_file = File(os.path.join(tmpdir, "code.zip"))
            create_zip(temp_file.path, ".", file_paths, arcname_prefix=arcname_prefix)
        else:
            temp_file = File(os.path.join(tmpdir, "code.tar.gz"))
            create_tar(temp_file.path, file_paths, arcname_prefix=arcname_prefix)

        if not basename:
            with temp_file.open("rb") as infile:
                basename = hash_stream(infile)

        code_file = File(get_code_scratch_file(scratch_prefix, basename, use_zip=use_zip))
        if not code_file.exists():
            temp_file.copy_to(code_file)

    return code_file



"""
Decorators that make libraries available to glue jobs.
"""

import importlib
import os
import site
import time
import zipfile

import s3fs

RDKIT_ZIP_LOCATION = "s3://insitro-user-robin/glue_scripts/rdkit_full_env.zip"


def enable_rdkit(zipfile: str = RDKIT_ZIP_LOCATION) -> None:
    """
    Installs the rdkit "binary zipfile" to user's site packages.
    Safe to call from Spark workers: will install on individual node
    and is thread and process safe.

    This CANNOT be a decorator as decorators aren't picklable and
    you might want to use rdkit in a Spark UDF, etc, which is pickled.
    """
    try:
        from rdkit import Chem  # noqa: F401

    except (ImportError, ModuleNotFoundError):
        print("Installing zipped module")
        install_zipped_module(zipfile)


def install_zipped_module(s3_path: str) -> None:
    """
    Installs a zip file from S3 to user's site packages.
    Thread and process safe- can be called from Spark.
    """

    # Try to get the "unzip lock"
    lockfile = os.path.join(str(site.USER_BASE), ".unzip_lock")
    try:
        lock_fd = os.open(lockfile, os.O_CREAT | os.O_EXCL)

        # Download the module from S3.
        s3 = s3fs.S3FileSystem()
        local_fn = "./tmp_zipmod.zip"
        s3.get(s3_path, local_fn)
        print("Downloaded zipfile")

        # Unpack the module to site packages.
        with zipfile.ZipFile(local_fn, mode="r") as zf:
            zf.extractall(site.USER_BASE)
        print("Extracted zipfile")

        # Release lock
        os.close(lock_fd)
        os.remove(lockfile)

    # If someone else has the lock, wait for them to install the module.
    except FileExistsError:
        while os.path.exists(lockfile):
            time.sleep(1.0)

    importlib.invalidate_caches()



import importlib
import typing
from typing import Any, Callable, Dict, Iterator, Optional, Type, cast

if typing.TYPE_CHECKING:
    from redun.scheduler import Job, Scheduler


class ExecutorError(Exception):
    pass


class Executor:
    """Note that most Executors should Register themselves with the method `register_executor`,
    below."""

    def __init__(
        self,
        name: str,
        scheduler: Optional["Scheduler"] = None,
        config=None,
    ):
        self.name = name
        self._scheduler = scheduler

    def set_scheduler(self, scheduler: "Scheduler") -> None:
        self._scheduler = scheduler

    def log(self, *messages: Any, **kwargs) -> None:
        """
        Display log message through Scheduler.
        """
        assert self._scheduler
        self._scheduler.log(f"Executor[{self.name}]:", *messages, **kwargs)

    def submit(self, job: "Job") -> None:
        """Execute the provided job.

        Implementations must provide results back to the scheduler by either calling `done_job` or
        `reject_job`."""
        assert self._scheduler
        self._scheduler.reject_job(
            job,
            ExecutorError("Executor {} does not support submitting tasks.".format(type(self))),
        )

    def submit_script(self, job: "Job") -> None:
        """Execute the provided script job.

        Implementations must provide results back to the scheduler by either calling `done_job` or
        `reject_job`."""

        assert self._scheduler
        self._scheduler.reject_job(
            job,
            ExecutorError(
                "Executor {} does not support submitting script tasks.".format(type(self))
            ),
        )

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    def scratch_root(self) -> str:
        raise NotImplementedError()


# Singleton executor registry.
_executor_classes: Dict[str, Type[Executor]] = {}


def get_executor_class(executor_name: str, required: bool = True) -> Optional[Type[Executor]]:
    """
    Get an Executor by name from the executor registry.

    Parameters
    ----------
    executor_name : str
        Name of executor class to retrieve.
    required : bool
        If True, raises error if executor is not registered.
        If False, None is returned for unknown executor name.
    """
    executor_class = _executor_classes.get(executor_name)
    if required and not executor_class:
        raise ExecutorError("Unknown executor {}".format(executor_name))
    return executor_class


def _register_executor(executor_name: str, executor_class: Type[Executor]) -> None:
    """
    Register an Executor class to be used by the scheduler.
    """
    _executor_classes[executor_name] = executor_class


def register_executor(executor_name: str) -> Callable:
    """
    Register an Executor class to be used by the scheduler.

    Note that registered classes are responsible for ensuring their modules get loaded, so this
    decorator is actually run. For example, this can be done by adding them to the redun
    module `__init__.py`.
    """

    def deco(executor_class: Type[Executor]):
        _register_executor(executor_name, executor_class)
        return executor_class

    return deco


def get_executors_from_config(executors_config: dict) -> Iterator[Executor]:
    """
    Instantiate executors defined in an executors config section.
    """
    for executor_name, executor_config in executors_config.items():
        executor_class = cast(Type[Executor], get_executor_class(executor_config["type"]))
        executor = executor_class(executor_name, config=executor_config)
        yield executor


def get_executor_from_config(executors_config: dict, executor_name: str) -> Executor:
    """Create and return the executor by name. Raise an error if it is not present."""
    for name, executor_config in executors_config.items():
        if name != executor_name:
            continue

        executor_class = cast(Type[Executor], get_executor_class(executor_config["type"]))
        executor = executor_class(executor_name, config=executor_config)
        return executor
    raise RuntimeError(f"Unknown Executor {executor_name}.")


def load_task_module(module_name: str, task_name: str) -> None:
    """
    Helper method that Executors may call to load a task's module code.

    Presently, the main benefit is this method provides better error handling.

    Args:
          module_name: Task module name.
          task_name: Task name
    """
    try:
        importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        # Append user-friendly error to original exception
        raise ModuleNotFoundError(
            f"Failed to find module {module_name} for task {task_name}. "
            "If this is remotely executed, ensure module is included in code packaging."
        ) from exc



from textwrap import dedent
from typing import Any, Callable, Dict, List, Sequence, Tuple, TypeVar, Union

from redun.expression import Expression, SchedulerExpression
from redun.promise import Promise
from redun.scheduler import Job, Scheduler, scheduler_task
from redun.task import PartialTask, Task, task

T = TypeVar("T")


@task(namespace="redun", version="1")
def compose_apply(tasks: Sequence[Task], *args: Any, **kwargs: Any) -> Any:
    """
    Helper function for applying a composition of Tasks to arguments.
    """
    rev_tasks = reversed(tasks)
    result = next(rev_tasks)(*args, **kwargs)
    for a_task in rev_tasks:
        result = a_task(result)
    return result


def compose(*tasks: Task) -> Task:
    """
    Compose Tasks together.
    """
    return compose_apply.partial(tasks)


@task(namespace="redun", version="1")
def identity(x: T) -> T:
    """
    Returns its input argument.
    """
    return x


def delay(x: T) -> PartialTask:
    """
    Delay the evaluation of a value x.

    The name `delay()` is inspired from other programming languages:
    - http://web.mit.edu/scheme_v9.2/doc/mit-scheme-ref/Promises.html
    - http://people.cs.aau.dk/~normark/prog3-03/html/notes/eval-order_themes-delay-stream-section.html  # noqa: E501
    - https://docs.racket-lang.org/reference/Delayed_Evaluation.html
    """
    # PartialTasks do not evaluate their arguments, so this acts similar to
    # a closure: `lambda: x`.
    return identity.partial(x)


def force(x: Task[Callable[[], T]]) -> T:
    """
    Force the evaluation of a delayed evaluation.

    The name `force()` is inspired from other programming languages:
    - http://web.mit.edu/scheme_v9.2/doc/mit-scheme-ref/Promises.html
    - http://people.cs.aau.dk/~normark/prog3-03/html/notes/eval-order_themes-delay-stream-section.html  # noqa: E501
    - https://docs.racket-lang.org/reference/Delayed_Evaluation.html
    """
    # delay() produces a Task of no arguments. To force evaluation we just need
    # to call the Task.
    return x()


@scheduler_task(namespace="redun")
def seq(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    exprs: Sequence[Any],
) -> Promise:
    """
    Evaluate the expressions serially.

    The name `seq()` is inspired from Haskell:
    - https://wiki.haskell.org/Seq
    """
    if not exprs:
        # No expressions, so return empty list.
        return Promise(lambda resolve, reject: resolve([]))

    result = []

    def then(pair):
        i, value = pair
        i += 1
        result.append(value)
        if i < len(exprs):
            return scheduler.evaluate((i, exprs[i]), parent_job=parent_job).then(then)
        else:
            return result

    return scheduler.evaluate((0, exprs[0]), parent_job=parent_job).then(then)


@task(namespace="redun", version="1")
def const(x: T, _: Any) -> T:
    """
    Returns the first argument unchanged (constant) and discards the second.

    This is useful for forcing execution of the second argument, but discarding
    it's value from downstream parts of the workflow.
    """
    return x


@task(namespace="redun", version="1")
def apply_func(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """
    Apply a Python function to possibly lazy arguments.

    The function `func` must be defined at the module-level.

    .. code-block:: python

        lazy_list = a_task_returning_a_list()
        lazy_length = apply_func(len)(lazy_list)
    """
    return func(*args, **kwargs)


def as_task(func: Callable[..., T]) -> PartialTask:
    """
    Transform a plain Python function into a redun Task.

    The function `func` must be defined at the module-level.

    .. code-block:: python

        assert as_task(max)(10, 20, 30, 25) == 30
    """
    return apply_func.partial(func)


@task(namespace="redun", version="1")
def eval_(code: str, *args: Any, **kwargs: Any) -> Any:
    """
    Evaluate `code` with `kwargs` as local variables.

    If a position argument is given, its value is assigned to local
    variables defined by `pos_args`.

    This can be useful for small manipulations of lazy values.

    .. code-block:: python

        records = task1()
        names = eval_("[record.name for record in records]", records=records)
        result = task2(names)
    """
    pos_args = kwargs.pop("pos_args", [])
    if len(pos_args) != len(args):
        raise TypeError(
            f"Wrong number of positional arguments given {len(args)}, expected {len(pos_args)}."
        )
    vars = dict(zip(pos_args, args))
    vars.update(kwargs)

    return eval(dedent(code), vars)


@scheduler_task(namespace="redun", version="1")
def map_(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    a_task: Task,
    values: Sequence[Any],
) -> Promise:
    """
    Map a task to a list of values, similar to `map(f, xs)`.
    """
    tasks = [a_task]

    # As an optimization, compose multiple maps into one.
    # e.g. map_(g, map_(f, xs)) == map_(compose(g, f), xs)
    while isinstance(values, SchedulerExpression) and values.task_name == "redun.map_":
        tasks.append(values.args[0])
        values = values.args[1]

    if len(tasks) == 1:
        [a_task] = tasks
    else:
        a_task = compose(*tasks)

    def then(a_task: Task) -> Promise:
        if isinstance(values, (list, tuple)):
            # Ready to perform parallel map.
            return scheduler.evaluate(
                [a_task(value) for value in values],
                parent_job=parent_job,
            )
        else:
            # Need to evaluate list first.
            return scheduler.evaluate(values, parent_job=parent_job).then(
                lambda values: scheduler.evaluate(
                    [a_task(value) for value in values],
                    parent_job=parent_job,
                )
            )

    # Evaluate task first, in case it's an expression.
    return scheduler.evaluate(a_task, parent_job=parent_job).then(then)


def starmap(
    a_task: Task[Callable[..., T]],
    kwargs: Union[List[Dict], Expression[List[Dict]]] = [],
) -> List[T]:
    """
    Map a task to a list of keyword arguments.
    """
    return map_(eval_.partial("a_task(**x)", pos_args=["x"], a_task=a_task), kwargs)


@task(namespace="redun", version="1")
def flatten(list_of_lists: Sequence[Sequence[T]]) -> List[T]:
    """
    Flatten a list of lists into a flat list.
    """
    return [value for lst in list_of_lists for value in lst]


@task(namespace="redun", version="1")
def flat_map(a_task: Task[Callable[..., List[T]]], values: List) -> List[T]:
    """
    Apply a task `a_task` on a sequence of `values` and flatten the result.
    """
    return flatten(map_(a_task, values))


@task(namespace="redun", version="1")
def zip_(*lists: List[T]) -> List[Tuple[T, ...]]:
    """
    Zips two or lists into a list of tuples.

    This is a task equivalent of zip().
    """
    return list(zip(*lists))



"""Add Tag models

Revision ID: cc4f663817b6
Revises: 647c510a77b1
Create Date: 2020-08-16 19:39:28.772684

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "cc4f663817b6"
down_revision = "cd2d53191748"
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == "postgresql":
        json_type = sa.dialects.postgresql.JSONB
    else:
        json_type = sa.String

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "tag",
        sa.Column("tag_hash", sa.String(length=40), nullable=False),
        sa.Column(
            "entity_type",
            sa.Enum(
                "Execution",
                "Job",
                "CallNode",
                "Task",
                "Value",
                "Null",
                name="tagentitytype",
            ),
            nullable=False,
        ),
        sa.Column("entity_id", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("value", json_type, nullable=False),
        sa.Column("is_current", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("tag_hash"),
    )
    op.create_index(op.f("ix_tag_entity_id"), "tag", ["entity_id"], unique=False)
    op.create_index(op.f("ix_tag_key"), "tag", ["key"], unique=False)
    op.create_index(op.f("ix_tag_value"), "tag", ["value"], unique=False)
    op.create_index(
        "ix_tag_tag_hash_current",
        "tag",
        ["tag_hash"],
        unique=True,
        postgresql_where=sa.text("is_current"),
        sqlite_where=sa.text("is_current"),
    )

    op.create_table(
        "tag_edit",
        sa.Column("parent_id", sa.String(), nullable=False),
        sa.Column("child_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["child_id"],
            ["tag.tag_hash"],
        ),
        sa.ForeignKeyConstraint(
            ["parent_id"],
            ["tag.tag_hash"],
        ),
        sa.PrimaryKeyConstraint("parent_id", "child_id"),
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("tag_edit")
    op.drop_table("tag")
    # ### end Alembic commands ###



import typing
from shlex import quote
from typing import Optional, cast

from redun.config import Config
from redun.executors.base import Executor, get_executor_from_config
from redun.expression import TaskExpression
from redun.logging import logger
from redun.scheduler import Execution, Job, Scheduler
from redun.scripting import script_task
from redun.task import hash_args_eval
from redun.value import get_type_registry


def launch_script(
    config: Config,
    script_command: typing.List[str],
    executor: Optional[Executor] = None,
    executor_name: Optional[str] = None,
    task_options: Optional[dict] = None,
    execution_id: Optional[str] = None,
) -> None:
    """
    Submit the provided script command to the executor, then exit.

    Use a local scheduler with the default config. This means we won't record the entry point,
    but we have no intention of being around long enough to record the results, so there's not
    much point.

    WARNING: This won't actually work on all executor types, such as the local ones. To work,
    the executor needs to be "fire and forget" for `submit_script`.

    Parameters
    ----------
    config : Config
        Config object containing executor information.
    script_command : List[str]
        The script command to run on the executor.
    executor : Optional[Executor]
        Optional Executor on which to run the script.  Must be set if executor_name is not set.
    executor_name : Optional[str]
        Name of the executor in the config.  Must be set if executor is not set.
    task_options : Optional[dict]
        Task options to pass to the script task.
    execution_id : Optional[str]
        If provided, use this execution id. This is only relevant for tagging.

    Returns
    -------
        None
    """

    if executor is None:
        assert executor_name is not None, "Must provide an executor by object or by name."
        executor = get_executor_from_config(config.get("executors", {}), executor_name)

    scheduler = Scheduler()
    execution = Execution(id=execution_id)
    executor.set_scheduler(scheduler)

    # Prepare command to execute within Executor.
    remote_run_command = " ".join(quote(arg) for arg in script_command)

    task_options = task_options or {}
    task_options["executor"] = executor_name

    # Setup job for inner run command.
    run_expr = cast(TaskExpression, script_task.options(**task_options)(remote_run_command))

    logger.info(f"Run within Executor {executor_name}: {remote_run_command}")

    # Submit directly to executor and immediately exit.
    job = Job(script_task, run_expr, execution=execution)
    script_args = ()
    script_kwargs = {"command": remote_run_command}
    job.eval_hash, job.args_hash = hash_args_eval(
        get_type_registry(), script_task, script_args, script_kwargs
    )
    job.args = script_args, script_kwargs

    # Submit job to executor.
    executor.submit_script(job)
    executor.stop()  # stop the monitor thread.



import sys
import typing
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from multiprocessing import get_context, get_start_method, set_start_method
from typing import Any, Callable, Optional, Tuple

from redun.config import create_config_section
from redun.executors.base import Executor, load_task_module, register_executor
from redun.scripting import exec_script, get_task_command
from redun.task import get_task_registry

if typing.TYPE_CHECKING:
    from redun.scheduler import Job, Scheduler


# Modes of local execution.
THREAD_MODE = "thread"
PROCESS_MODE = "process"


def exec_task(mode: str, module_name: str, task_fullname: str, args: Tuple, kwargs: dict) -> Any:
    """
    Execute a task in the new process.
    """
    load_task_module(module_name, task_fullname)
    task = get_task_registry().get(task_fullname)
    return task.func(*args, **kwargs)


def exec_script_task(
    mode: str, module_name: str, task_fullname: str, args: Tuple, kwargs: dict
) -> bytes:
    """
    Execute a script task from the task registry.
    """
    load_task_module(module_name, task_fullname)
    task = get_task_registry().get(task_fullname)
    command = get_task_command(task, args, kwargs)
    return exec_script(command)


@register_executor("local")
class LocalExecutor(Executor):
    """
    A redun Executor for running jobs locally using a thread or process pool.
    """

    # Available local executor modes.
    MODES = [THREAD_MODE, PROCESS_MODE]
    _OLD2NEW_MODES = {
        "threads": THREAD_MODE,
        "processes": PROCESS_MODE,
    }
    START_METHODS = ["fork", "spawn", "forkserver"]
    # start_method fork is not reliable on Mac OS X. So we use forkserver as
    # a safe common default.
    # https://bugs.python.org/issue33725
    DEFAULT_START_METHOD = "forkserver"

    def __init__(
        self,
        name: str,
        scheduler: Optional["Scheduler"] = None,
        config=None,
        mode: str = THREAD_MODE,
    ):
        super().__init__(name, scheduler=scheduler)

        # Parse config.
        if not config:
            config = create_config_section()

        self.max_workers = config.getint("max_workers", 20)
        self.mode = config.get("mode", mode)
        # Autoconvert deprecated modes.
        self.mode = self._OLD2NEW_MODES.get(self.mode, self.mode)
        self._scratch_root = config.get("scratch", "/tmp/redun")
        assert self.mode in self.MODES, f"Unknown mode: {self.mode}"

        self.start_method = config.get("start_method", self.DEFAULT_START_METHOD)
        assert (
            self.start_method in self.START_METHODS
        ), f"Unknown start_method: {self.start_method}"

        # Pools.
        self._thread_executor: Optional[ThreadPoolExecutor] = None
        self._process_executor: Optional[ProcessPoolExecutor] = None

    def _start(self, mode: str) -> None:
        """
        Start pool on first Job submission.
        """
        if mode == THREAD_MODE and not self._thread_executor:
            self._thread_executor = ThreadPoolExecutor(max_workers=self.max_workers)

        if mode == PROCESS_MODE and not self._process_executor:
            if sys.version_info < (3, 7):
                # https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor
                # Changed in version 3.7: The mp_context argument was added to allow users
                # to control the start_method for worker processes created by the pool.
                # https://github.com/python/cpython/blob/3.6/Lib/multiprocessing/context.py#L240
                if get_start_method(allow_none=True) is None:
                    set_start_method(self.start_method)
                self._process_executor = ProcessPoolExecutor(max_workers=self.max_workers)
            else:
                self._process_executor = ProcessPoolExecutor(
                    max_workers=self.max_workers,
                    mp_context=get_context(self.start_method),
                )

    def stop(self) -> None:
        """
        Stop Executor pools.
        """
        if self._thread_executor:
            self._thread_executor.shutdown()
            self._thread_executor = None

        if self._process_executor:
            # Shutdown causes problems on python3.8
            # https://bugs.python.org/issue39995
            if (sys.version_info.major, sys.version_info.minor) != (3, 8):
                self._process_executor.shutdown()
            self._process_executor = None

    def _submit(self, exec_func: Callable, job: "Job") -> None:
        mode = job.get_option("mode", self.mode)
        if mode not in (THREAD_MODE, PROCESS_MODE):
            raise ValueError(f"Unknown mode: {mode}")

        # Ensure pool are started.
        self._start(mode)

        # Determine pool executor.
        executor = (
            self._thread_executor
            if mode == THREAD_MODE
            else self._process_executor
            if mode == PROCESS_MODE
            else None
        )
        if not executor:
            raise ValueError('Unknown LocalExecutor.mode "{}"'.format(mode))

        # Run job in a new thread or process.

        def on_done(future):
            try:
                self._scheduler.done_job(job, future.result())
            except Exception as error:
                self._scheduler.reject_job(job, error)

        assert job.args
        args, kwargs = job.args
        executor.submit(
            exec_func, mode, job.task.load_module, job.task.fullname, args, kwargs
        ).add_done_callback(on_done)

    def submit(self, job: "Job") -> None:
        assert not job.task.script
        self._submit(exec_task, job)

    def submit_script(self, job: "Job") -> None:
        assert job.task.script
        self._submit(exec_script_task, job)

    def scratch_root(self) -> str:
        return self._scratch_root



from typing import Any, Optional, TypeVar, cast

from redun.expression import SchedulerExpression
from redun.promise import Promise
from redun.scheduler import Job, Scheduler
from redun.task import scheduler_task

T = TypeVar("T")


def get_context_value(context: dict, var_path: str, default: T) -> T:
    """
    Returns a value from a context dict using a dot-separated path (e.g. 'foo.bar.baz').

    Parameters
    ----------
    context : dict
        A redun context dict.
    var_path : str
        A dot-separated path for a variable to fetch from the current context.
    default : T
        A default value to return if the context variable is not defined.
    """
    parts = var_path.split(".")
    value = context
    try:
        for part in parts:
            if not isinstance(value, dict):
                return default
            value = value[part]
    except KeyError:
        return default
    return cast(T, value)


@scheduler_task(namespace="redun")
def get_context(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    var_path: str,
    default: Optional[Any] = None,
) -> Promise:
    """
    Returns a value from the current context.

    Parameters
    ----------
    var_path : str
        A dot-separated path for a variable to fetch from the current context.
    default : Optional[Any]
        A default value to return if the context variable is not defined. None by default.
    """
    return scheduler.evaluate(parent_job.get_context(), parent_job=parent_job).then(
        lambda context: get_context_value(context, var_path, default)
    )



import importlib
import os
import pkgutil
import re
import shutil
import tempfile
import time
from contextlib import contextmanager
from functools import wraps
from inspect import getmembers, getmodule, isclass, isfunction, ismethod, ismodule
from itertools import zip_longest
from typing import IO, Any, Callable, Dict, Iterator, List, NamedTuple, Optional, Type
from unittest.mock import patch

import sqlalchemy.event
from aiobotocore.response import StreamingBody
from botocore.awsrequest import AWSResponse
from moto import mock_s3 as _mock_s3
from moto.core.botocore_stubber import MockRawResponse

# This should be public, and will be in v2.0. Until then, live with the private import.
from urllib3._collections import HTTPHeaderDict

from redun import Scheduler
from redun.file import (
    FileSystem,
    GSFileSystem,
    S3FileSystem,
    get_filesystem_class,
    get_proto,
)

"""
Current moto (1.3.16) is not fully compatible with the latest s3fs (2021.11.1),
because of s3fs's use of aiobotocore. Here we define a small monkey patch
to make moto compatible enough for our needs. If future versions of moto
gain full support, these patches can be removed.
"""


class AsyncMockedRawResponse(MockRawResponse):
    """
    Wraps a moto MockRawResponse in order to provide an async read method.
    """

    def __init__(self, response):
        self._response = response
        self._content = b""

    async def read(self, *args, **kwargs):
        # Provide an async version of the read method.
        return self._sync_read(*args, **kwargs)

    def _sync_read(self, *args, **kwargs):
        data = self._response.read(*args, **kwargs)
        self._content += data
        return data

    def stream(self, **kwargs):
        # We override stream in order to call the original synchronous read.
        contents = self._sync_read()
        while contents:
            yield contents
            contents = self._sync_read()

    @property
    def content(self):
        return self


async def ident(x):
    return x


class AsyncBytes(bytes):
    def __await__(self):
        return ident(bytes(self)).__await__()


class PatchedStreamingBody(StreamingBody):
    def read(self, amt=None):
        # This is a read method that can be used sync or async.
        return AsyncBytes(self._sync_read(amt))


class PatchedAWSResponse(AWSResponse):
    """
    Subclasses AWSRespnse in order to provide an async read and raw headers.

    Inspiration: https://github.com/aio-libs/aiobotocore/issues/755#issuecomment-807325931
    """

    def __init__(self, url, status_code, headers, raw):
        async_raw = PatchedStreamingBody(
            AsyncMockedRawResponse(raw), headers.get("content-length")
        )
        super().__init__(url, status_code, headers, async_raw)

    @property
    def raw_headers(self):
        # Raw headers are expected to be an iterable of pairs of bytes. We can
        # reconstruct the raw headers from the parsed ones.
        return [
            (key.encode("utf-8"), str(value).encode("utf-8"))
            for key, value in self.headers.items()
        ]

    async def read(self):
        return self.text

    @property
    async def content(self):
        return self.raw._content


def convert_to_response_dict_patch(http_response, operation_model):
    """
    Patched version of botocore.endpoint.convert_to_repsonse
    """
    response_dict = {
        # botocore converts keys to str, so make sure that they are in
        # the expected case. See detailed discussion here:
        # https://github.com/aio-libs/aiobotocore/pull/116
        # aiohttp's CIMultiDict camel cases the headers :(
        "headers": HTTPHeaderDict(
            {k.decode("utf-8").lower(): v.decode("utf-8") for k, v in http_response.raw_headers}
        ),
        "status_code": http_response.status_code,
        "context": {
            "operation_name": operation_model.name,
        },
    }
    if response_dict["status_code"] >= 300:
        response_dict["body"] = http_response.raw._sync_read()
    elif operation_model.has_event_stream_output:
        response_dict["body"] = http_response.raw
    elif operation_model.has_streaming_output:
        length = response_dict["headers"].get("content-length")
        response_dict["body"] = PatchedStreamingBody(http_response.raw, length)
    else:
        response_dict["body"] = http_response.raw._sync_read()
    return response_dict


async def async_convert_to_response_dict_patch(http_response, operation_model):
    """
    Patched version of aiobotocore.endpoint.convert_to_repsonse
    """
    return convert_to_response_dict_patch(http_response, operation_model)


def patch_aws(func: Callable) -> Callable:
    """
    Fix botocore and moto to work together.

    This patch is needed until moto is fully compatiable with latest botocore.
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        # Apply patches.
        with patch("botocore.awsrequest.AWSResponse", PatchedAWSResponse), patch(
            "moto.core.botocore_stubber.AWSResponse", PatchedAWSResponse
        ), patch(
            "botocore.endpoint.convert_to_response_dict", convert_to_response_dict_patch
        ), patch(
            "aiobotocore.endpoint.convert_to_response_dict",
            async_convert_to_response_dict_patch,
        ):
            return func(*args, **kwargs)

    return wrapped


def mock_s3(func: Callable) -> Callable:
    """
    This is redun's wrapped mock_s3 with applies patch_aws.

    All uses of mock_s3 should use this one in order to make sure the moto
    compatibility fixes are in place.
    """
    return patch_aws(_mock_s3(func))


def get_filesystem_class_mock(
    proto: Optional[str] = None, url: Optional[str] = None
) -> Type[FileSystem]:
    """
    Mock redun filesystem lookup.
    """
    if not proto:
        assert url, "Must give url or proto as argument."
        proto = get_proto(url)
    if proto == "gs":
        return GSFileSystemMock
    else:
        return get_filesystem_class(proto, url)


class GSFileSystemMock(GSFileSystem):
    """
    This class mocks the Google Cloud Storage file system by proxying to s3 mocked by moto.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._s3fs = S3FileSystem()

    def convert_path(self, path: str) -> str:
        """
        Convert path from gs to s3.
        """
        if path.startswith("gs://"):
            return "s3://" + path[5:]
        else:
            return path

    def unconvert_path(self, path: str) -> str:
        """
        Convert path from s3 to gs.
        """
        if path.startswith("s3://"):
            return "gs://" + path[5:]
        else:
            return path

    def _open(self, path: str, mode: str, **kwargs: Any) -> IO:
        """
        Private open method for subclasses to implement.
        """
        path = self.convert_path(path)
        return self._s3fs._open(path, mode, **kwargs)

    def exists(self, path: str) -> bool:
        """
        Returns True if path exists on filesystem.
        """
        return self._s3fs.exists(self.convert_path(path))

    def remove(self, path: str) -> None:
        """
        Delete a path from the filesystem.
        """
        return self._s3fs.remove(self.convert_path(path))

    def mkdir(self, path: str) -> None:
        """
        Creates the directory in the filesystem.
        """
        return self._s3fs.mkdir(self.convert_path(path))

    def rmdir(self, path: str, recursive: bool = False) -> None:
        """
        Removes a directory from the filesystem.
        If `recursive`, removes all contents of the directory.
        Otherwise, raises OSError on non-empty directories
        """
        return self._s3fs.rmdir(self.convert_path(path))

    def get_hash(self, path: str) -> str:
        """
        Return a hash for the file at path.
        """
        return self._s3fs.get_hash(self.convert_path(path))

    def copy(self, src_path: str, dest_path: str) -> None:
        """
        Copy a file from src_path to dest_path.
        """
        return self._s3fs.copy(self.convert_path(src_path), self.convert_path(dest_path))

    def glob(self, pattern: str) -> List[str]:
        """
        Returns filenames matching pattern.
        """
        paths = self._s3fs.glob(self.convert_path(pattern))
        return [self.unconvert_path(path) for path in paths]

    def isfile(self, path: str) -> bool:
        """
        Returns True if path is a file.
        """
        return self._s3fs.isfile(self.convert_path(path))

    def isdir(self, path: str) -> bool:
        """
        Returns True if path is a directory.
        """
        return self._s3fs.isdir(self.convert_path(path))

    def filesize(self, path: str) -> int:
        """
        Returns file size of path in bytes.
        """
        return self._s3fs.filesize(self.convert_path(path))


def clean_dir(path: str) -> None:
    """
    Ensure path exists and is an empty directory.
    """
    if os.path.exists(path):
        shutil.rmtree(path)
    os.makedirs(path)


def get_test_file(filename: str) -> str:
    """
    Returns a file from test_data.
    """
    basedir = os.path.dirname(__file__)
    return os.path.join(basedir, filename)


def use_tempdir(func: Callable) -> Callable:
    """
    Run function within a temporary directory.
    """

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        with tempfile.TemporaryDirectory() as tmpdir:
            original_dir = os.getcwd()
            os.chdir(tmpdir)

            try:
                result = func(*args, **kwargs)
            finally:
                os.chdir(original_dir)
        return result

    return wrap


def assert_match_lines(patterns: List[str], lines: List[str]) -> None:
    """
    Asserts whether `lines` match `patterns`.
    """
    assert len(patterns) == len(lines)
    for pattern, line in zip(patterns, lines):
        assert re.fullmatch(pattern, line), f"Could not match `{pattern}` to `{line}`"


def assert_match_text(pattern: str, text: str, wildcard: str = "*"):
    """
    Assert whether two strings are equal using wildcards.
    """
    for i, (a, b) in enumerate(zip_longest(pattern, text)):
        if a != b and a != wildcard:
            assert False, "mismatch on character {}: '{}' != '{}'".format(
                i, pattern[: i + 1], text[: i + 1]
            )


def wait_until(cond: Callable[[], bool], interval: float = 0.02, timeout: float = 2.0) -> None:
    """
    Wait until `cond()` is True or timeout is exceeded.
    """
    start = time.time()
    while not cond():
        if time.time() - start > timeout:
            raise RuntimeError("Timeout")
        time.sleep(interval)


class MatchEnv:
    """

    An environment for generating Match objects.
    """

    def __init__(self) -> None:
        self.vars: Dict[str, Any] = {}

    def match(self, *args, **kwargs) -> "Match":
        kwargs["env"] = self
        return Match(*args, **kwargs)


class Match:
    """
    Helper for asserting values have particular properties (types, etc).
    """

    def __init__(
        self,
        type: Optional[Type] = None,
        var: Optional[str] = None,
        regex: Optional[str] = None,
        any: bool = True,
        env: Optional[MatchEnv] = None,
    ):
        self.any = any
        self.type = type
        self.var = var
        self.regex = regex
        self.env = env

    def __repr__(self) -> str:
        if self.var:
            return "Match(var={})".format(self.var)
        elif self.type:
            return "Match(type={})".format(self.type.__name__)
        elif self.regex:
            return "Match(regex={})".format(self.regex)
        elif self.any:
            return "Match(any=True)"
        else:
            return "Match()"

    def __eq__(self, other: Any) -> bool:
        if self.env and self.var:
            # First instance of var will always return True.
            # Second instance of var has to match previous value.
            expected = self.env.vars.setdefault(self.var, other)
            if expected != other:
                return False

        if self.type:
            return isinstance(other, self.type)

        elif self.regex:
            return bool(re.fullmatch(self.regex, other))

        else:
            return self.any


class QueryStats(NamedTuple):
    """
    Stats for a recorded SQLAlchemy query.
    """

    statement: str
    parameters: tuple
    duration: float


@contextmanager
def listen_queries(engine: Any) -> Iterator[List[QueryStats]]:
    """
    Context for capturing SQLAlchemy queries.

    .. code-block:: python

        with listen_queries(engine) as queries:
            result = session.query(Model).filter(...)
            # More SQLAlchemy queries...

        # queries now has a list of statement and parameter tuples.
        assert len(queries) == 2
    """
    queries = []
    cursors = {}

    def before(conn, cursor, statement, parameters, context, executemany):
        cursors[cursor] = time.time()

    def after(conn, cursor, statement, parameters, context, executemany):
        duration = time.time() - cursors.pop(cursor)
        queries.append(QueryStats(statement, parameters, duration))

    sqlalchemy.event.listen(engine, "before_cursor_execute", before)
    sqlalchemy.event.listen(engine, "after_cursor_execute", after)

    yield queries

    sqlalchemy.event.remove(engine, "before_cursor_execute", before)
    sqlalchemy.event.remove(engine, "after_cursor_execute", after)


def import_all_modules(pkg):
    """Import (almost) all modules within a package.

    Ignores explicitly marked modules.
    """
    ignored_modules = (
        "redun.backends.db.alembic.env",
        "redun.visualize",
    )  # https://stackoverflow.com/a/52575218
    modules = []
    for _, module_name, is_pkg in pkgutil.iter_modules(pkg.__path__):
        full_name = f"{pkg.__name__}.{module_name}"
        if full_name in ignored_modules:
            continue

        module = importlib.import_module(full_name)
        if is_pkg:
            modules.extend(import_all_modules(module))
        else:
            modules.append(module)

    return modules


def get_docstring_owners_in_module(module):
    """Get all functions, classes and their methods defined within a python module.

    Returns
    -------
    docstring_owners : set
        Set of functions, classes and methods
    """
    assert ismodule(module), f"Passed {module.__name__} which is not a module."

    def is_valid(obj):
        if getmodule(obj) == module:
            if ismethod(obj) or isfunction(obj):
                return not obj.__name__.startswith("_")
            if isclass(obj):
                return True
        return False

    to_check = {obj for _, obj in getmembers(module) if is_valid(obj)}
    docstring_owners = set()
    seen = set()

    while to_check:
        candidate = to_check.pop()
        if candidate in seen:
            continue

        if isfunction(candidate) or ismethod(candidate):
            docstring_owners.add(candidate)

        if isclass(candidate):
            to_check.update({obj for _, obj in getmembers(candidate) if is_valid(obj)})

        seen.add(candidate)
    return docstring_owners


def docstring_owner_pretty_name(docstring_owner):
    return ".".join((docstring_owner.__module__, docstring_owner.__qualname__))


def mock_scheduler():
    """
    Returns a scheduler with mocks for job completion.
    """
    # Setup scheduler callbacks.
    scheduler = Scheduler()

    scheduler.job_results = {}
    scheduler.job_errors = {}

    def done_job(job, result, job_tags=[]):
        job.job_tags.extend(job_tags)
        scheduler.job_results[job.id] = result

    def reject_job(job, error, error_traceback=None, job_tags=[]):
        if job:
            job.job_tags.extend(job_tags)
            scheduler.job_errors[job.id] = error
        else:
            # Scheduler error, reraise it.
            scheduler.job_errors[None] = error
            raise error

    def batch_wait(job_ids):
        while not all(
            job_id in scheduler.job_results or job_id in scheduler.job_errors for job_id in job_ids
        ):
            time.sleep(0.1)

    scheduler.done_job = done_job
    scheduler.reject_job = reject_job
    scheduler.batch_wait = batch_wait

    return scheduler



# Utility routines for k8s API interactions
# This library contains standardized routines for interacting with k8s through its API
# It uses the Official Python client library for kubernetes:
# https://github.com/kubernetes-client/python
from base64 import b64encode
from typing import Any, Dict, List, Optional, Tuple

from kubernetes import client, config
from kubernetes.client.exceptions import ApiException
from kubernetes.config import ConfigException

from redun.logging import logger

DEFAULT_JOB_PREFIX = "redun-job"


class K8SClient:
    """
    This class manages multiple k8s clients and their config loading.
    """

    def __init__(self) -> None:
        self._is_loaded = False
        self._core: Optional[client.CoreV1Api] = None
        self._custom_objects: Optional[client.CustomObjectsApi] = None
        self._batch: Optional[client.BatchV1Api] = None
        self._rbac: Optional[client.RbacAuthorizationV1Api] = None

    def load_config(self) -> None:
        """
        Loads k8s config if not loaded.
        """
        if not self._is_loaded:
            try:
                config.load_kube_config()
            except ConfigException:
                logger.warn(
                    "config.load_kube_config() failed. "
                    "Resorting to config.load_incluster_config()."
                )
                config.load_incluster_config()
            self._is_loaded = True

    def version(self) -> Tuple[int, int]:
        """
        Returns an API client support k8s version API.

        https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/VersionApi.md
        """
        self.load_config()
        version_info = client.VersionApi().get_code()
        major = int(version_info.major.replace("+", ""))
        minor = int(version_info.minor.replace("+", ""))
        return major, minor

    @property
    def core(self) -> client.CoreV1Api:
        """
        Returns an API client support k8s core API.

        https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/CoreV1Api.md
        """
        if not self._core:
            self.load_config()
            self._core = client.CoreV1Api()
        return self._core

    @property
    def custom_objects(self) -> client.CustomObjectsApi:
        """
        Returns an API client support k8s custom_objects API.

        https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/CustomObjectsApi.md
        """
        if not self._custom_objects:
            self.load_config()
            self._custom_objects = client.CustomObjectsApi()
        return self._custom_objects

    @property
    def batch(self) -> client.BatchV1Api:
        """
        Returns an API client supporting k8s batch API.

        https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/BatchV1Api.md
        """
        if not self._batch:
            self.load_config()
            self._batch = client.BatchV1Api()
        return self._batch

    @property
    def rbac(self) -> client.RbacAuthorizationV1Api:
        """
        Returns an API client support k8s RbacAuthorizationV1 API.

        https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/RbacAuthorizationV1Api.md
        """
        if not self._rbac:
            self.load_config()
            self._rbac = client.RbacAuthorizationV1Api()
        return self._rbac


def delete_k8s_secret(k8s_client: K8SClient, secret_name: str, namespace: str) -> None:
    """
    Deletes a k8s Secret. If Secret doesn't exist, it does not throw an error.
    """
    try:
        k8s_client.core.delete_namespaced_secret(secret_name, namespace)
    except ApiException as error:
        if error.status != 404:
            # If secret doesn't exist, then this is a no-op.
            raise


def create_k8s_secret(
    k8s_client: K8SClient,
    secret_name: str,
    namespace: str,
    secret_data: dict,
    secret_type: str = "Opaque",
) -> Any:
    """
    Creates a k8s Secret from a dict of key-value pairs.
    """
    # Build secret.
    metadata = {"name": secret_name, "namespace": namespace}
    data = {
        key: b64encode(value.encode("utf8")).decode("utf8") for key, value in secret_data.items()
    }
    body = client.V1Secret("v1", data, False, "Secret", metadata, type=secret_type)

    # Create secret.
    try:
        return k8s_client.core.create_namespaced_secret(namespace, body)
    except ApiException as error:
        if error.status == 409:
            # Secret already exists, just patch it.
            return k8s_client.core.replace_namespaced_secret(secret_name, namespace, body)
        raise


def create_resources(
    requests: Optional[dict] = None, limits: Optional[dict] = None
) -> client.V1ResourceRequirements:
    """
    Creates resource limits for k8s pods.

    https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1ResourceRequirements.md
    """
    resources = client.V1ResourceRequirements()

    if requests is None:
        resources.requests = {}
    else:
        resources.requests = requests

    if limits is None:
        resources.limits = {}
    else:
        resources.limits = limits

    return resources


def create_job_object(
    name: str = DEFAULT_JOB_PREFIX,
    image: str = "bash",
    command: List[str] = ["false"],
    resources: Optional[client.V1ResourceRequirements] = None,
    timeout: Optional[int] = None,
    labels: Optional[dict] = None,
    uid: Optional[str] = None,
    retries: int = 1,
    service_account_name: Optional[str] = "default",
    annotations: Optional[Dict[str, str]] = None,
    secret_name: Optional[str] = None,
) -> client.V1Job:
    """Creates a job object for redun job.
    https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Job.md
    Also creates necessary sub-objects"""

    # Container environment variables.
    env = []

    # Forward AWS secrets to the container environment variables.
    aws_env_keys = ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN"]
    if secret_name:
        env.extend(
            [
                {
                    "name": key,
                    "valueFrom": {
                        "secretKeyRef": {
                            "name": secret_name,
                            "key": key,
                            "optional": True,
                        },
                    },
                }
                for key in aws_env_keys
            ]
        )

    container = client.V1Container(name=name, image=image, args=command, env=env)

    if resources is None:
        container.resources = create_resources()
    else:
        container.resources = resources
    pod_spec = client.V1PodSpec(
        service_account_name=service_account_name,
        restart_policy="Never",
        image_pull_secrets=[{"name": "regcred"}],
        containers=[container],
    )

    # Create and configurate a pod template spec section
    template = client.V1PodTemplateSpec(metadata=client.V1ObjectMeta(), spec=pod_spec)

    # Create the spec of job deployment
    spec = client.V1JobSpec(
        template=template,
        backoff_limit=retries,
        # ttl_seconds_after_finished=100,
    )
    if timeout:
        spec.active_deadline_seconds = timeout

    if labels is None:
        labels = {}

    # Instantiate the job object
    return client.V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=client.V1ObjectMeta(annotations=annotations, name=name, labels=labels, uid=uid),
        spec=spec,
    )


def create_namespace(k8s_client: K8SClient, namespace: str) -> None:
    """Create a k8s namespace"""
    try:
        k8s_client.core.create_namespace(
            client.V1Namespace(metadata=client.V1ObjectMeta(name=namespace))
        )
    except client.exceptions.ApiException as error:
        if error.status == 409 and error.reason == "Conflict":
            pass
        else:
            logger.error("Unexpected exception creating namespace", error.body)
            raise


def annotate_service_account(
    k8s_client: K8SClient, namespace: str, sa_name: str, annotations_to_append: dict
) -> None:
    """Annotate a service account with additional annotations

    Args:
        namespace: Namespace wherein service account resides.
        sa_name: Service account name.
        annotations_to_append: Dict of annotations to append.
    """
    sa = k8s_client.core.read_namespaced_service_account(sa_name, namespace)

    existing_annotations = sa.metadata.annotations or {}
    existing_annotations.update(annotations_to_append)
    sa.metadata.annotations = existing_annotations
    try:
        k8s_client.core.replace_namespaced_service_account(sa_name, namespace, sa)
        logger.info(
            f"ServiceAccount: {sa_name} in namespace: {namespace} annotated with "
            f"{existing_annotations}"
        )
    except client.exceptions.ApiException as error:
        logger.error("Unexpected exception annotating SA: ", error.body)
        raise


def delete_namespace(k8s_client: K8SClient, namespace: str) -> None:
    """Delete a k8s namespace"""
    try:
        k8s_client.core.delete_namespace(name=namespace)
    except client.exceptions.ApiException as error:
        logger.error("Unexpected exception deleting namespace", error.body)
        raise


def create_job(k8s_client: K8SClient, job: client.V1Job, namespace: str) -> client.V1Job:
    """
    Creates an actual k8s job.
    """
    try:
        return k8s_client.batch.create_namespaced_job(body=job, namespace=namespace)
    except client.exceptions.ApiException as error:
        if error.status == 409 and error.reason == "Conflict":
            # Job already exsists, return it.
            return k8s_client.batch.read_namespaced_job(job.metadata.name, namespace=namespace)
        else:
            logger.error("Error submitting k8s job:", error.body)
            raise


def delete_job(k8s_client: K8SClient, name: str, namespace: str) -> Any:
    """
    Deletes an existing k8s job.
    """
    return k8s_client.batch.delete_namespaced_job(
        name=name,
        namespace=namespace,
        body=client.V1DeleteOptions(propagation_policy="Background"),
    )



"""
Dataflow visualization.

An upstream dataflow visualization explains the derivation of a value. Take
for example this dataflow visualization of the derivation of a VCF file
from a bioinformatic analysis:

```
value = File(path=sample4.vcf, hash=********)

value <-- <********> call_variants(bam, ref_genome)
  bam        = <********> File(path=sample4.bam, hash=********)
  ref_genome = <********> File(path=ref_genome, hash=********)

bam <-- argument of <********> call_variants_all(bams, ref_genome)
    <-- <********> align_reads_all(fastqs, ref_genome)
    <-- <********> align_reads(fastq, ref_genome)
  fastq      = <********> File(path=sample4.fastq, hash=********)
  ref_genome = <********> File(path=ref_genome, hash=********)

fastq <-- argument of <********> align_reads_all(fastqs, ref_genome)
      <-- argument of <********> main(fastqs, ref_genome)
      <-- origin

ref_genome <-- argument of <********> align_reads_all(fastqs, ref_genome)
           <-- argument of <********> main(fastqs, ref_genome)
           <-- origin
```

Hash values are indicated by * above.  For reference, here is what the
workflow might have been:

```
    @task()
    def align_reads(fastq: File, ref_genome: File) -> File:
        reads = cast(str, fastq.read())
        ref = cast(str, ref_genome.read())
        bam = File(fastq.path.replace("fastq", "bam"))
        bam.write("align({}, {})".format(reads, ref))
        return bam

    @task()
    def call_variants(bam: File, ref_genome: File) -> File:
        align = cast(str, bam.read())
        ref = cast(str, ref_genome.read())
        vcf = File(bam.path.replace("bam", "vcf"))
        vcf.write("calls({}, {})".format(align, ref))
        return vcf

    @task()
    def align_reads_all(fastqs: List[File], ref_genome: File):
        bams = [align_reads(fastq, ref_genome) for fastq in fastqs]
        return bams

    @task()
    def call_variants_all(bams: List[File], ref_genome: File):
        vcfs = [call_variants(bam, ref_genome) for bam in bams]
        return vcfs

    @task()
    def main(fastqs: List[File], ref_genome: File):
        bams = align_reads_all(fastqs, ref_genome)
        vcfs = call_variants_all(bams, ref_genome)
        return vcfs
```

A dataflow visualization consists of a series of paragraphs called
"dataflow sections" that describe how one of the values is derived. Here
is the section for the `bam` value:

```
bam <-- argument of <********> call_variants_all(bams, ref_genome)
    <-- <********> align_reads_all(fastqs, ref_genome)
    <-- <********> align_reads(fastq, ref_genome_2)
  fastq      = <********> File(path=sample4.fastq, hash=********)
  ref_genome = <********> File(path=ref_genome, hash=********)
```

A section is made of three clauses: assignment, routing, and arguments.

The assignment clause indicates which CallNode produced this value:

```
bam <-- argument of <********> call_variants_all(bams, ref_genome)
```

Routing clauses, if present, describe a series of additional CallNodes
that "route" the value by passing via arguments from parent CallNode to child
CallNode, or by results from child CallNode to parent CallNode.

```
    <-- result of <********> align_reads_all(fastqs, ref_genome)
    <-- <********> align_reads(fastq, ref_genome_2)
```

Argument clauses define the value for each argument in the final CallNode.

```
  fastq      = <********> File(path=sample4.fastq, hash=********)
  ref_genome = <********> File(path=ref_genome, hash=********)
```

To build this visualization, the following strategy is used:
- Given a starting value (e.g. a VCF file in the example above), walk the
  CallGraph backwards (i.e. upstream) to determine relevant nodes. These are
  call DataflowNodes, which are connected by DataflowEdges.
- DataflowEdges are then grouped into sections.
- Each section is then reorganized into a DataflowSectionDOM. A DataflowDOM
  is the collection of DataflowSectionDOMs. The DOM(document object model) is
  an intermediate representation that can be rendered in multiple ways.
- Once a DataflowDOM is created, it can either be rendered into a textual
  format, or serialized into JSON for the web frontend.
"""

import ast
import re
from collections import defaultdict
from enum import Enum
from itertools import chain
from textwrap import dedent
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from redun.backends.db import Argument, CallNode, RedunBackendDb, Task, Value
from redun.utils import assert_never, trim_string

T = TypeVar("T")

REDUN_INTERNAL_TASKS = {
    "redun.postprocess_script",
    "redun.script",
}


def iter_unique(items: Iterable[T], key: Callable[[T], Any] = lambda x: x) -> Iterator[T]:
    """
    Iterate through unique items.
    """
    seen: Set[T] = set()
    for item in items:
        item_key = key(item)
        if item_key not in seen:
            yield item
            seen.add(item_key)


class ArgumentValue(NamedTuple):
    """
    A DataflowNode used for tracing one subvalue in an argument.
    """

    argument: Argument
    value: Value


class CallNodeValue(NamedTuple):
    """
    A DataflowNode used for tracing one subvalue of a CallNode result.
    """

    call_node: CallNode
    value: Value


# There are several kinds of DataflowNodes.
DataflowNode = Union[ArgumentValue, CallNodeValue, CallNode, Value]


class DataflowEdge(NamedTuple):
    """
    An edge in a Dataflow graph.
    """

    src: DataflowNode
    dest: Optional[DataflowNode]


# A grouping of DataflowEdges that are displayed as one "paragraph".
DataflowSection = List[DataflowEdge]


class DataflowSectionKind(Enum):
    """
    Each dataflow section describes either a task call or a data manipulation.
    """

    CALL = "call"
    DATA = "data"


class DataflowAssign(NamedTuple):
    """
    The assignment clause in a Dataflow DOM.
    """

    var_name: str
    prefix: str
    node_display: str
    node: Optional[DataflowNode]


class DataflowRouting(NamedTuple):
    """
    A routing clause in a Dataflow DOM.
    """

    prefix: str
    node_display: str
    node: Optional[DataflowNode]


class DataflowArg(NamedTuple):
    """
    An argument clause in a Dataflow DOM.
    """

    var_name: str
    value: Value


class DataflowSectionDOM(NamedTuple):
    """
    A section in Dataflow DOM.
    """

    assign: DataflowAssign
    routing: List[DataflowRouting]
    args: List[DataflowArg]


# The top-level Dataflow DOM.
DataflowDOM = Iterable[DataflowSectionDOM]


def get_task_args(task: Task) -> List[str]:
    """
    Returns list of argument names of a Task. Raises a SyntaxError if the task source code is not
    properly formatted.
    """
    # Since we don't currently record the name of positional arguments,
    # we have to infer them from the source code.
    code = ast.parse(dedent(task.source))

    if not isinstance(code.body[0], ast.FunctionDef):
        raise SyntaxError("Source code is not a properly formatted function.")

    # Type ignore is needed since the AST lib does seem to use proper types
    # everywhere.
    return [arg.arg for arg in code.body[0].args.args]


def make_var_name(var_name_base: str, name2var: Dict[str, DataflowNode], suffix: int = 2) -> str:
    """
    Generate a new variable using a unique suffix (e.g. myvar_2).
    """
    # Strip numerical suffix.
    var_name_base = re.sub(r"_\d+$", "", var_name_base)

    if var_name_base not in name2var:
        # This variable name is already unique.
        return var_name_base

    # Search increase suffixes until we find a unique variable name.
    while True:
        new_var_name = var_name_base + "_" + str(suffix)
        if new_var_name not in name2var:
            return new_var_name
        suffix += 1


def get_default_arg_name(pos: int) -> str:
    """
    Generate the default name for argument.
    """
    return f"arg{pos}"


class DataflowVars:
    """
    Manages variable names for nodes in a dataflow.
    """

    def __init__(self) -> None:
        self.var2name: Dict[DataflowNode, str] = {}
        self.name2var: Dict[str, DataflowNode] = {}
        self.task2args: Dict[Task, List[str]] = {}

    def __getitem__(self, node: DataflowNode) -> str:
        """
        Get a variable name for a DataflowNode.
        """
        return self.var2name[node]

    def __setitem__(self, node: DataflowNode, var_name: str) -> str:
        """
        Set a new variable name for a DataflowNode.
        """
        self.var2name[node] = var_name
        self.name2var[var_name] = node
        return var_name

    def __contains__(self, node: DataflowNode) -> bool:
        """
        Returns True if node has a variable name.
        """
        return node in self.var2name

    def get_task_args(self, task: Task) -> List[str]:
        """
        Returns the parameter names of a Task.
        """
        args = self.task2args.get(task)
        if not args:
            # Cache task arg names.
            # TODO: Properly handle variadic args.
            try:
                args = self.task2args[task] = get_task_args(task)
            except SyntaxError:
                # The argument names cannot be properly parsed.
                return []
        return args

    def new_var_name(
        self, node: DataflowNode, base_var_name: Optional[str] = None
    ) -> Tuple[str, Optional[str]]:
        """
        Get or create a new variable name for a DataflowNode.
        """
        var_name = self.var2name.get(node)
        if var_name:
            # Node is already named.
            return var_name, None

        if not base_var_name:
            # Autogenerate base var name from ArgumentValue.
            assert isinstance(node, ArgumentValue)
            argument, value = node

            # Determine new variable name.
            if argument.arg_key:
                base_var_name = argument.arg_key
            else:
                arg_names = self.get_task_args(argument.call_node.task)
                if not arg_names:
                    # Use default argument names.
                    base_var_name = get_default_arg_name(argument.arg_position)
                else:
                    base_var_name = arg_names[argument.arg_position]

        # Ensure variable name is unique.
        var_name = make_var_name(base_var_name, self.name2var)
        self[node] = var_name
        return var_name, base_var_name


def walk_dataflow_value(backend: RedunBackendDb, value: Value) -> Iterator[DataflowEdge]:
    """
    Iterates through the edges in the upstream dataflow graph of a Value.
    """

    # Find upstream CallNodes.
    # A value can be produced by many CallNodes and it can be a subvalue of
    # a value produced from many CallNodes.
    call_nodes = set(value.results) | {
        call_node for parent in value.parents for call_node in parent.results
    }
    call_nodes = {call_node for call_node in call_nodes if not is_internal_task(call_node.task)}

    def walk_parents(node: CallNode, seen: set) -> Iterator[CallNode]:
        for parent in node.parents:
            if parent not in seen:
                yield parent
                seen.add(parent)
                yield from walk_parents(parent, seen)

    # Determine which CallNodes are just routing CallNodes.
    # A routing CallNode is an upstream CallNode that is also an ancestor
    # of another upstream CallNode.
    seen: Set[CallNode] = set()
    routing_call_nodes = set()
    for call_node in call_nodes:
        for ancestor in walk_parents(call_node, seen):
            if ancestor in call_nodes:
                routing_call_nodes.add(ancestor)
                break

    # Determine originating CallNodes (upstream and non-routing).
    originating_call_nodes = call_nodes - routing_call_nodes

    # Prefer the most recent CallNodes.
    start_time_call_nodes = [
        (job.start_time, call_node)
        for call_node in originating_call_nodes
        for job in call_node.jobs
    ]
    max_start_time = max((start_time for start_time, _ in start_time_call_nodes), default=None)
    upstream_call_node = next(
        (
            call_node
            for start_time, call_node in start_time_call_nodes
            if start_time == max_start_time
        ),
        None,
    )

    # Emit Value-CallNode edges in dataflow.
    if upstream_call_node:
        yield DataflowEdge(value, upstream_call_node)
    else:
        yield DataflowEdge(value, None)


def get_callnode_arguments(call_node: CallNode) -> List[Argument]:
    """
    Returns a CallNode's arguments in sorted order.
    """
    pos_args = []
    kw_args = []
    for arg in call_node.arguments:
        if arg.arg_position is not None:
            pos_args.append(arg)
        else:
            kw_args.append(arg)
    pos_args.sort(key=lambda arg: arg.arg_position)
    kw_args.sort(key=lambda arg: arg.arg_key)
    return pos_args + kw_args


def walk_dataflow_callnode(backend: RedunBackendDb, call_node: CallNode) -> Iterator[DataflowEdge]:
    """
    Iterates through the upstream Arguments of a CallNode.
    """
    # Emit CallNode-ArgumentValue edges in dataflow.
    # Reversing the arguments will lead to the traversal to be in original
    # argument order, which is nicer for display.
    arguments = get_callnode_arguments(call_node)
    if arguments:
        for argument in arguments:
            yield DataflowEdge(call_node, ArgumentValue(argument, argument.value))
    else:
        # There are no arguments, this is an origin of the dataflow.
        yield DataflowEdge(call_node, None)


def is_internal_task(task: Task) -> bool:
    """
    Returns True if task is an internal redun task.

    We skip such tasks in the dataflow to avoid clutter.
    """
    return task.fullname in REDUN_INTERNAL_TASKS


def walk_dataflow_callnode_value(
    backend: RedunBackendDb, call_node_value: CallNodeValue
) -> Iterator[DataflowEdge]:
    """
    Iterates through the upstream dataflow edges of a CallNodeValue.

    The edges either go deeper to the child CallNodes or stay with this CallNode.
    """
    call_node, value = call_node_value

    def is_subvalue(query_value: Value, target_value: Value) -> bool:
        value_hashes = chain(
            [target_value.value_hash],
            (child.value_hash for child in target_value.children),
        )
        return query_value.value_hash in value_hashes

    # Prefer the flow from children of call_node over call_node.
    child_matches = [
        child_node
        for child_node in call_node.children
        if not is_internal_task(child_node.task) and is_subvalue(value, child_node.value)
    ]
    if len(child_matches) == 1:
        # There is one obvious child CallNode that produced this value.
        # Follow the dataflow through this child CallNode.
        [child_node] = child_matches
        yield DataflowEdge(call_node_value, CallNodeValue(child_node, value))
    else:
        # Otherwise, we follow the dataflow for the whole CallNode.
        yield DataflowEdge(call_node_value, call_node)


def walk_dataflow_argument_value(
    backend: RedunBackendDb, argument_value: ArgumentValue
) -> Iterator[DataflowEdge]:
    """
    Iterates through the upstream dataflow edges of an ArgumentValue.

    Edge types:
    - ArgumentValue <-- CallNodeValue:
      - Value came from the result of a CallNode.
    - ArgumentValue <-- ArgumentValue
      - Value came from argument of parent CallNode, i.e. argument-to-argument routing.
    - ArgumentValue <-- Origin
      - Value came directly from user or task.
    """
    argument, value = argument_value
    is_terminal = True

    # Determine the most recent common parent CallNode of all the upstream CallNodes.
    call_node_parents = [set(call_node.parents) for call_node in argument.upstream]
    call_node_parents.append(set(argument.call_node.parents))
    context_call_node = next(
        iter(
            sorted(
                set.intersection(*call_node_parents),
                key=lambda call_node: call_node.timestamp,
                reverse=True,
            )
        ),
        None,
    )

    # Process upstream CallNodes in a consistent order (call_order).
    if context_call_node:
        # Use reverse order during emission to get correct order during display.
        upstream_order = sorted(
            [
                (call_node, edge.call_order)
                for call_node in argument.upstream
                for edge in call_node.parent_edges
                if edge.parent_node == context_call_node
            ],
            key=lambda pair: pair[1],
            reverse=True,
        )
        upstream = [call_node for call_node, _ in upstream_order]
    else:
        # Fallback if we can't determine context call node.
        upstream = argument.upstream

    # Emit upstream sibling CallNodes.
    for call_node in iter_unique(upstream):
        # Order subvalues for consistency.
        subvalues = sorted(call_node.value.children, key=lambda child: child.value_hash)
        result_values = iter_unique(chain([call_node.value], subvalues))
        match = False
        for result_value in result_values:
            if value.value_hash == result_value.value_hash:
                is_terminal = False
                match = True
                yield DataflowEdge(
                    argument_value,
                    CallNodeValue(call_node, result_value),
                )

        if not match:
            # Default to emitting the whole result of the CallNode.
            is_terminal = False
            yield DataflowEdge(
                argument_value,
                CallNodeValue(call_node, call_node.value),
            )

    # Emit upstream argument from parent CallNode.
    # Prefer most recent parent.
    parent_call_nodes = sorted(
        argument.call_node.parents, key=lambda parent: parent.timestamp, reverse=True
    )
    for parent_call_node in parent_call_nodes[:1]:
        for parent_argument in parent_call_node.arguments:
            parent_values = chain([parent_argument.value], parent_argument.value.children)
            for parent_value in parent_values:
                if value.value_hash == parent_value.value_hash:
                    is_terminal = False
                    yield DataflowEdge(
                        argument_value,
                        ArgumentValue(parent_argument, value),
                    )

    # Emit terminal origin value.
    if is_terminal:
        yield DataflowEdge(argument_value, None)


def walk_dataflow_node(backend: RedunBackendDb, node: DataflowNode) -> Iterator[DataflowEdge]:
    """
    Iterates through the upstream dataflow edges of a any DataflowNode.
    """
    if isinstance(node, Value):
        return walk_dataflow_value(backend, node)

    elif isinstance(node, CallNode):
        return walk_dataflow_callnode(backend, node)

    elif isinstance(node, ArgumentValue):
        return walk_dataflow_argument_value(backend, node)

    elif isinstance(node, CallNodeValue):
        return walk_dataflow_callnode_value(backend, node)

    else:
        assert_never(node)


def walk_dataflow(backend: RedunBackendDb, init_node: DataflowNode) -> Iterator[DataflowEdge]:
    """
    Iterate through all the upstream dataflow edges of a 'node' in the CallGraph.

    A 'node' can be a Value, CallNode, CallNodeValue, or an ArgumentValue.
    """
    # Perform depth-first traversal.
    queue: List[DataflowNode] = [init_node]
    seen: Set[DataflowNode] = set()

    while queue:
        node: DataflowNode = queue.pop()
        child_edges = list(walk_dataflow_node(backend, node))
        yield from child_edges

        # Reverse edges before pushing on to stack to maintain sibling order.
        for edge in reversed(child_edges):
            node2 = edge.dest
            if node2 is None:
                # Terminal node of type 'Origin'.
                continue

            # Determine whether dest node is unique and should be added to queue.
            if node2 not in seen:
                queue.append(node2)
                seen.add(node2)


def get_section_edge_type(edge: DataflowEdge) -> str:
    """
    Classifies a DataflowEdge.
    """
    src, dest = edge
    if (
        isinstance(src, ArgumentValue)
        and isinstance(dest, CallNodeValue)
        and src.value.value_hash != dest.value.value_hash
    ):
        return "data"

    elif isinstance(src, CallNode) and isinstance(dest, ArgumentValue):
        return "call_arg"

    elif isinstance(src, (Value, CallNodeValue)) and isinstance(dest, CallNode):
        return "call_result"

    elif isinstance(src, CallNodeValue) and isinstance(dest, CallNodeValue):
        return "call_result_routing"

    elif isinstance(src, ArgumentValue) and isinstance(dest, CallNodeValue):
        return "call_arg_result_routing"

    elif isinstance(src, ArgumentValue) and isinstance(dest, ArgumentValue):
        return "call_arg_routing"

    elif isinstance(src, CallNode) and dest is None:
        return "call_origin"

    elif isinstance(src, ArgumentValue) and dest is None:
        return "call_arg_origin"

    elif isinstance(src, Value) and dest is None:
        return "call_value_origin"

    else:
        raise AssertionError(f"Unknown edge type {edge}")


def toposort_edges(edges: Iterable[DataflowEdge]) -> Iterator[DataflowEdge]:
    """
    Topologically sort DataflowEdges in depth-first order.
    """

    # Compute indegree.
    indegrees: Dict[DataflowNode, int] = defaultdict(int)
    src2edge: Dict[DataflowNode, List[DataflowEdge]] = defaultdict(
        cast(Callable[[], List[DataflowEdge]], list)
    )
    for edge in edges:
        src2edge[edge.src].append(edge)

        # Ensure every node is present in indegrees, including roots.
        indegrees[edge.src]

        if edge.dest:
            indegrees[edge.dest] += 1

    # Initialize queue with roots.
    queue: List[DataflowNode] = [node for node, degree in indegrees.items() if degree == 0]

    while queue:
        node = queue.pop()
        yield from src2edge[node]

        # Reverse edges before pushing on stack in order to maintain sibling order.
        for edge in reversed(src2edge[node]):
            if edge.dest:
                indegrees[edge.dest] -= 1
                if indegrees[edge.dest] == 0:
                    # All parents have been visited, we can enqueue dest.
                    queue.append(edge.dest)


def rewrite_call_node_merges(edges: List[DataflowEdge]) -> List[DataflowEdge]:
    """
    Rewrites dataflow graphs to enforce one CallNodeValue per CallNode.

    This function identifies CallNode that have multilple parent CallNodeValues
    like this:

      ArgumentValue --> CallNodeValue --\
                                         V
                                         CallNode (merge_node)
                                         ^
      ArgumentValue --> CallNodeValue --/

    and rewrite them to unify the CallNodeValues like this:

      ArgumentValue --\
                       V
                       CallNodeValue --> CallNode
                       ^
      ArgumentValue --/
    """

    # Build graph dict.
    src2edges: Dict[Optional[DataflowNode], List[DataflowEdge]] = defaultdict(list)
    dest2edges = defaultdict(list)
    nodes = []
    for edge in edges:
        nodes.append(edge.src)
        src2edges[edge.src].append(edge)
        if edge.dest:
            dest2edges[edge.dest].append(edge)

    # Find merge nodes.
    merge_nodes = [
        node for node, edges in dest2edges.items() if isinstance(node, CallNode) and len(edges) > 1
    ]

    # Rewrite CallNode merge nodes.
    for merge_node in merge_nodes:
        # Find relevant edges and nodes.
        cv_cn_edges = dest2edges[merge_node]
        call_node_values = [edge.src for edge in cv_cn_edges]
        upstream_edges = [
            edge for call_node_value in call_node_values for edge in dest2edges[call_node_value]
        ]
        upstream_nodes = list(iter_unique(edge.src for edge in upstream_edges))
        old_edges = set(cv_cn_edges) | set(upstream_edges)

        # Create new unified CallNodeValue from call_node_values.
        unified_node = CallNodeValue(call_node=merge_node, value=merge_node.value)

        # Create new edges.
        new_edges = [
            DataflowEdge(
                src=upstream_node,
                dest=unified_node,
            )
            for upstream_node in upstream_nodes
        ] + [DataflowEdge(src=unified_node, dest=merge_node)]

        # Remove old edges from edges.
        edges2 = [edge for edge in edges if edge not in old_edges]

        # To keep edges in traversal order, insert new edges right before
        # the first appearance of the merge_node.
        insert_index = min(i for i, edge in enumerate(edges2) if edge.src == merge_node)
        edges = edges2[:insert_index] + new_edges + edges2[insert_index:]

    return edges


def iter_subsections(section: DataflowSection) -> Iterator[DataflowSection]:
    """
    Determines if a section should be broken down into small sections.

    In real life dataflows, there are some cases where the dataflow merges
    such that the structure is a DAG, not just a tree. These merges represent a
    value that was passed to two or more different tasks and then their outputs
    eventually combine again, either into a single Value like a list or as arguments
    into a common task. For example, `value` is a merge node in the upstream
    dataflow of `result`.

      value = task0()
      output1 = task1(a=value)
      output2 = task2(b=value)
      result = task3(c=output1, d=output2)

    The upstream dataflow graph of `result` is:

      Value(result)
       |
       V
      CallNode(task3) ---------------\
       |                             |
       V                             V
      ArgumentValue(task3, key=a) ArgumentValue(task3, key=b)
       |                             |
       V                             V
      CallNode(task1)             CallNode(task2)
       |                             |
       V                             V
      ArgumentValue(task1, key=c) ArgumentValue(task2, key=d)
       |                             |
       V                             |
      CallNodeValue(task0, value) <--/
       |
       V
      CallNode(task0)
       |
       V
      Origin

    The function `iter_dataflow_section()` will break this graph right after
    every the `CallNode --> ArgumentValue` edge, resulting in three sections.
    One of those sections will have a "merge node", `CallNode(task0)`:

      ArgumentValue(task1, key=c) ArgumentValue(task2, key=d)
       |                             |
       V                             |
      CallNodeValue(task0, value) <--/
       |
       V
      CallNode(task0)
       |
       V
      Origin

    This function will further break this section into subsections like this:

    Subsection 1:

      ArgumentValue(task1, key=c)
       |
       V
      CallNodeValue(task0, value)

    Subsection 2:

      ArgumentValue(task2, key=d)
       |
       V
      CallNodeValue(task0, value)

    Subsection 3:

      CallNodeValue(task0, value)
       |
       V
      CallNode(task0)
       |
       V
      Origin

    Ultimately these three subsections get rendered as:

      c <-- c_2

      d <-- c_2

      c_2 <-- task()
          <-- origin
    """

    # Build graph dict.
    src2edges: Dict[Optional[DataflowNode], List[DataflowEdge]] = defaultdict(list)
    dest2edges = defaultdict(list)
    nodes = []
    for edge in section:
        nodes.append(edge.src)
        src2edges[edge.src].append(edge)
        if edge.dest:
            dest2edges[edge.dest].append(edge)

    # Find roots. Maintain order of appearance in section.
    roots = [node for node in iter_unique(nodes) if len(dest2edges[node]) == 0]

    # Find merge nodes.
    merge_nodes = [node for node, edges in dest2edges.items() if len(edges) > 1]

    if not merge_nodes:
        # No merge nodes. Keep section as is.
        yield section
        return

    # Determine subsections.
    subsection: DataflowSection = []
    node: Optional[DataflowNode]
    for node in roots + merge_nodes:
        while True:
            next_edges = src2edges[node]
            if len(next_edges) != 1:
                # We have hit the end of the section.
                # There are either no more edges, or we hitting the arguments.
                subsection.extend(next_edges)
                yield subsection
                subsection = []
                break

            # Path should always be linear.
            [edge] = next_edges
            subsection.append(edge)

            if edge.dest not in merge_nodes:
                # Follow edge.
                node = edge.dest
            else:
                # We have hit the merge node, stop.
                yield subsection
                subsection = []
                break


def iter_dataflow_sections(
    dataflow_edges: Iterable[DataflowEdge],
) -> Iterator[Tuple[DataflowSectionKind, DataflowSection]]:
    """
    Yields dataflow sections from an iterable of dataflow edges.

    A dataflow section is a group of edges representing one 'paragraph' in a
    dataflow display.

      value <-- <1234abcd> call_node(arg1, arg2)
            <-- result of <2345abcd> call_node2(arg3, arg4)
        arg3 = <3456abcd> 'hello_world'
        arg4 = <4567abcd> File('foo.txt')
    """
    node2call_section: Dict[DataflowNode, List[Tuple[int, DataflowEdge]]] = {}
    node2data_section: Dict[DataflowNode, List[Tuple[int, DataflowEdge]]] = defaultdict(
        cast(Callable[[], List[Tuple[int, DataflowEdge]]], list)
    )

    new_vars: Set[ArgumentValue] = set()
    section: List[Tuple[int, DataflowEdge]]

    edges = toposort_edges(dataflow_edges)
    edge_list = rewrite_call_node_merges(list(edges))

    # Group dataflow edges into display sections.
    # Retain the appearance order of each edge so we can sort sections later.
    for i, edge in enumerate(edge_list):
        edge_type = get_section_edge_type(edge)

        if edge_type == "data":
            # Edge is a data section edge.
            node2data_section[edge.src].append((i, edge))
            continue

        if edge_type == "call_arg":
            # New variable.
            assert isinstance(edge.dest, ArgumentValue)
            new_vars.add(edge.dest)

        if edge.src in new_vars:
            # src is a new variable so we start a new section.
            section = [(i, edge)]

        else:
            # Get or create section associated with src and add this edge.
            if edge.src not in node2call_section:
                section = node2call_section[edge.src] = []
            else:
                section = node2call_section[edge.src]
            section.append((i, edge))

        # Get section associated with dest and union with src section.
        if edge.dest:
            if edge.dest not in node2call_section:
                # dest_section same as src.
                node2call_section[edge.dest] = section
            else:
                # Union dest_section with section.
                # Although this extend might have dups, I use iter_unique to clean it up.
                dest_section = node2call_section[edge.dest]
                dest_section.extend(section)
                node2call_section[edge.src] = dest_section

    # Get unique sections.
    call_sections = iter_unique(node2call_section.values(), key=id)
    data_sections = node2data_section.values()

    def get_order_section(
        int_edges: List[Tuple[int, DataflowEdge]],
    ) -> Tuple[int, List[DataflowEdge]]:
        """
        Returns a tuple of section appearance order and the section.

        The appearance order of a section is the maximum order of its edges.
        We also clean up any duplicate edges that may have been added to the section.
        """
        ints = []
        section = []
        for i, edge in int_edges:
            ints.append(i)
            section.append(edge)
        return (max(ints), list(iter_unique(section)))

    # Label each section with its type ("call" or "data") and determine section order.
    sections = [
        (DataflowSectionKind.CALL,) + get_order_section(int_edges) for int_edges in call_sections
    ]
    sections.extend(
        (DataflowSectionKind.DATA,) + get_order_section(int_edges) for int_edges in data_sections
    )

    # Sort sections.
    sections = sorted(sections, key=lambda row: row[1])

    # Yield sections.
    for kind, _, section2 in sections:
        if kind == DataflowSectionKind.CALL:
            # If there is path merging, then we will emit multiple sections.
            for subsection in iter_subsections(section2):
                yield (kind, subsection)
        else:
            yield (kind, section2)


def make_section_dom(
    section: DataflowSection,
    dataflow_vars: DataflowVars,
    new_varname: str = "value",
) -> DataflowSectionDOM:
    """
    Returns DOM for a dataflow section.
    """
    # Determine assign information from first edge in section.
    src, assign_node = section[0]
    if isinstance(src, Value):
        # Start new variable.
        assign_var_name = dataflow_vars[src] = new_varname

    elif isinstance(src, (ArgumentValue, CallNodeValue)):
        # Value should be named already.
        assign_var_name = dataflow_vars[src]

    else:
        assert not isinstance(src, CallNode)
        assert_never(src)

    routing_defs: List[Optional[DataflowNode]] = []
    arg_defs: List[Tuple[str, Value]] = []
    renames: Dict[str, str] = {}

    # Compute whether this section ends with a variable.
    is_var2var = isinstance(assign_node, (ArgumentValue, CallNodeValue))
    last_routing_node = None

    # Process remaining edges.
    for src, dest in section[1:]:
        if isinstance(src, CallNode) and isinstance(dest, ArgumentValue):
            # Argument definition edge.
            var_name, base_var_name = dataflow_vars.new_var_name(dest)
            if base_var_name:
                renames[base_var_name] = var_name
            arg_defs.append((var_name, dest.value))

        else:
            # Routing edge.
            last_routing_node = dest
            if isinstance(dest, CallNode):
                is_var2var = False

            # Skip unnecessary routing edge, ignore.
            if not (isinstance(src, CallNodeValue) and isinstance(dest, CallNode)):
                routing_defs.append(dest)

    # The last routing def or assign (if there are no routing defs) needs
    # to rename its variables to be unique.
    last_line = len(routing_defs)

    # Create assign clause.
    if is_var2var and not routing_defs:
        # Assignment is a var2var.
        assert assign_node
        # Name the last node after the first node, if it doesn't have a name yet.
        base_name = dataflow_vars[section[0].src]
        var_name, _ = dataflow_vars.new_var_name(assign_node, base_name)
        dom_assign = DataflowAssign(assign_var_name, "", var_name, assign_node)
    else:
        prefix, node_display = display_node(assign_node, renames if last_line == 0 else {})
        dom_assign = DataflowAssign(assign_var_name, prefix, node_display, assign_node)

    # Create routing clauses.
    dom_routing: List[DataflowRouting] = []
    for i, dest in enumerate(routing_defs, 1):
        prefix, node_display = display_node(dest, renames if i == last_line else {})
        dom_routing.append(DataflowRouting(prefix, node_display, dest))

    # If last routing node is a Value, as opposed to a CallNode, then this was a
    # merge node and we should end with a new variable.
    if is_var2var and isinstance(last_routing_node, (ArgumentValue, CallNodeValue)):
        # This is a merge node, give it a variable name.
        assert not arg_defs

        # Name the last node after the first node, if it doesn't have a name yet.
        base_name = dataflow_vars[section[0].src]
        var_name, _ = dataflow_vars.new_var_name(last_routing_node, base_name)
        dom_routing.append(DataflowRouting("", var_name, last_routing_node))

    # Create argument definitions.
    dom_args: List[DataflowArg] = [DataflowArg(var_name, var) for var_name, var in arg_defs]

    return DataflowSectionDOM(dom_assign, dom_routing, dom_args)


def make_data_section_dom(
    section: DataflowSection,
    dataflow_vars: DataflowVars,
    new_varname: str = "value",
) -> DataflowSectionDOM:
    """
    Returns a DOM for a data section.

    A data section describes how one value was constructed from several subvalues.
    For example this dataflow section:

      parent_value <-- derives from
        subvalue1 = File(path='file1')
        subvalue2 = File(path='file2')

    corresponds to this kind of code:

      subvalue1 = task1()
      subvalue2 = task2()
      parent_value = [subvalue1, subvalue2]
      task3(parent_value)
    """
    downstream_nodes: Set[DataflowNode] = set()
    upstream_nodes: List[CallNodeValue] = []

    for edge in section:
        assert isinstance(edge.dest, CallNodeValue)
        downstream_nodes.add(edge.src)
        upstream_nodes.append(edge.dest)

    # Determine assign clause.
    # We only support one downstream node currently.
    [downstream_node] = downstream_nodes
    assign_var_name = dataflow_vars[downstream_node]

    # Determine arg clauses.
    args = []
    for upstream_node in iter_unique(upstream_nodes):
        call_node, value = upstream_node

        # Ensure variable name is unique.
        base_var_name = upstream_node.call_node.task.name + "_result"
        new_var_name, _ = dataflow_vars.new_var_name(upstream_node, base_var_name)
        args.append(DataflowArg(new_var_name, value))

    return DataflowSectionDOM(
        assign=DataflowAssign(
            var_name=assign_var_name, prefix="", node_display="derives from", node=None
        ),
        routing=[],
        args=args,
    )


def make_dataflow_dom(
    dataflow_edges: Iterable[DataflowEdge], new_varname: str = "value"
) -> Iterable[DataflowSectionDOM]:
    """
    Yields dataflow section DOMs from an iterable of dataflow edges.

    It also performs variable renaming to give every value a unique variable name.
    """
    dataflow_vars = DataflowVars()

    for kind, section in iter_dataflow_sections(dataflow_edges):
        if kind == DataflowSectionKind.CALL:
            yield make_section_dom(section, dataflow_vars, new_varname=new_varname)

        elif kind == DataflowSectionKind.DATA:
            yield make_data_section_dom(section, dataflow_vars, new_varname=new_varname)

        else:
            raise NotImplementedError(f"Unknown kind '{kind}'.")


def get_dataflow_call_node(node: Optional[DataflowNode]) -> Optional[CallNode]:
    """
    Returns the CallNode for a DataflowNode.
    """
    if isinstance(node, CallNode):
        return node

    elif isinstance(node, ArgumentValue):
        return node.argument.call_node

    elif isinstance(node, CallNodeValue):
        return node.call_node

    elif node is None:
        return None

    else:
        assert_never(node)


def get_node_hash(node: Optional[DataflowNode]) -> Optional[str]:
    """
    Formats hash for a DataflowNode.
    """
    if isinstance(node, Value):
        return node.value_hash

    call_node = get_dataflow_call_node(node)
    if call_node:
        return call_node.call_hash

    return None


def display_node(node: Optional[DataflowNode], renames: Dict[str, str]) -> Tuple[str, str]:
    """
    Formats a dataflow node to a string.
    """
    if isinstance(node, CallNode):
        return ("", display_call_node(node, renames))

    elif isinstance(node, ArgumentValue):
        return (
            "argument of",
            display_call_node(node.argument.call_node, renames),
        )

    elif isinstance(node, CallNodeValue):
        return ("", display_call_node(node.call_node, renames))

    elif node is None:
        return ("", "origin")

    else:
        assert_never(node)


def display_call_node(call_node: CallNode, renames: Dict[str, str]) -> str:
    """
    Formats a CallNode to a string.
    """
    try:
        arg_names = get_task_args(call_node.task)
    except SyntaxError:
        arg_names = [get_default_arg_name(i) for i in range(len(call_node.arguments))]
    args = [renames.get(arg, arg) for arg in arg_names]
    return "{task_name}({args})".format(task_name=call_node.task.name, args=", ".join(args))


def display_value(value: Value) -> str:
    """
    Format a Value to a string.
    """
    return trim_string(repr(value.preview))


def display_hash(node: Optional[DataflowNode]) -> str:
    """
    Formats hash for a DataflowNode.
    """
    node_hash = get_node_hash(node)
    if node_hash:
        return "<{}> ".format(node_hash[:8])
    else:
        return ""


def display_section(dom: DataflowSectionDOM) -> Iterator[str]:
    """
    Yields lines for displaying a dataflow section DOM.
    """

    # Display assign line.
    assign = dom.assign
    yield "{var_name} <-- {prefix}{value_hash}{value}".format(
        var_name=assign.var_name,
        prefix=assign.prefix + " " if assign.prefix else "",
        value_hash=display_hash(assign.node),
        value=assign.node_display,
    )

    indent = len(assign.var_name) + 1

    # Display routing.
    for routing_def in dom.routing:
        yield "{indent}<-- {prefix}{node_hash}{node}".format(
            indent=" " * indent,
            prefix=routing_def.prefix + " " if routing_def.prefix else "",
            node_hash=display_hash(routing_def.node),
            node=routing_def.node_display,
        )

    # Display argument definitions.
    if dom.args:
        max_var_len = max(len(arg.var_name) for arg in dom.args)
        for arg in dom.args:
            yield "  {var_name}{padding} = <{value_hash}> {var}".format(
                var_name=arg.var_name,
                padding=" " * (max_var_len - len(arg.var_name)),
                value_hash=arg.value.value_hash[:8],
                var=display_value(arg.value),
            )


def display_dataflow(dom: DataflowDOM) -> Iterator[str]:
    """
    Yields for lines displaying a dataflow DOM.
    """
    for dom_section in dom:
        yield from display_section(dom_section)
        yield ""


def serialize_section(dom: DataflowSectionDOM) -> dict:
    """
    Serialize a DataflowSectionDOM to JSON.
    """
    # Serialize assignment.
    assign_hash: Optional[str]
    if isinstance(dom.assign.node, Value):
        assign_hash = dom.assign.node.value_hash
        assign_job_id = None
        assign_execution_id = None
    elif dom.assign.node:
        call_node = get_dataflow_call_node(dom.assign.node)
        assert call_node
        assign_hash = call_node.call_hash
        # Get latest job of call_node.
        assign_job = sorted(call_node.jobs, key=lambda job: job.start_time, reverse=True)[0]
        assign_job_id = assign_job.id
        assign_execution_id = assign_job.execution.id
    else:
        assign_hash = None
        assign_job_id = None
        assign_execution_id = None

    # Serialize routing.
    routing = []
    for routing_def in dom.routing:
        call_node = get_dataflow_call_node(routing_def.node)
        if call_node:
            call_hash: Optional[str] = call_node.call_hash
            # Get latest job of call_node.
            job = sorted(call_node.jobs, key=lambda job: job.start_time, reverse=True)[0]
            job_id: Optional[str] = job.id
            execution_id: Optional[str] = job.execution.id
        else:
            call_hash = None
            job_id = None
            execution_id = None

        routing.append(
            {
                "prefix": routing_def.prefix,
                "hash": call_hash,
                "display": routing_def.node_display,
                "job_id": job_id,
                "execution_id": execution_id,
            }
        )

    # Serialize arguments.
    args = [
        {
            "var_name": arg.var_name,
            "hash": arg.value.value_hash,
            "value_display": display_value(arg.value),
        }
        for arg in dom.args
    ]

    return {
        "assign": {
            "var_name": dom.assign.var_name,
            "prefix": dom.assign.prefix,
            "display": dom.assign.node_display,
            "hash": assign_hash,
            "job_id": assign_job_id,
            "execution_id": assign_execution_id,
        },
        "routing": routing,
        "args": args,
    }


def serialize_dataflow(dom: DataflowDOM) -> Iterator[dict]:
    """
    Serialize DataflowDOM to JSON.
    """
    for dom_section in dom:
        yield serialize_section(dom_section)



import json
from code import InteractiveInterpreter
from collections import defaultdict
from typing import Any, Dict, List, Optional, cast

from rich.console import RenderableType
from rich.style import Style
from rich.text import Text
from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import DataTable, Footer, Input, Static

from redun.backends.db import CallNode, Execution, Job, Tag, Task, Value
from redun.cli import format_timedelta, format_timestamp
from redun.console.utils import format_job, get_links, style_status
from redun.scheduler import Job as SchedulerJob

NULL = object()


class RedunHeader(Static):
    """
    Header used on all Screen to display the current CLI command (e.g. argv).
    """

    CSS_PATH = "style.css"

    def __init__(self, title: str, **kwargs):
        super().__init__("[bold]redun console[/bold] " + title, **kwargs)

    def update(self, title: str) -> None:
        super().update("[bold]redun console[/bold] " + title)


class PageLabel(Widget):
    page = reactive(1)

    def __init__(self, page: int):
        super().__init__()
        self.page = page

    def render(self) -> RenderableType:
        return Text(f" Page {self.page} ", style=Style(color="black", bgcolor="green"))


class RedunFooter(Footer):
    """
    Footer used on Screens that use pagination.
    """

    page = reactive(1)

    def __init__(self, page: int):
        super().__init__()
        self.page = page

    def compose(self) -> ComposeResult:
        yield from super().compose()
        self.styles.grid_size_columns += 1
        yield PageLabel(self.page)

    def watch_page(self) -> None:
        for page_label in self.query(PageLabel):
            page_label.page = self.page


class Table(DataTable):
    """
    Generic DataTable that uses Enter key to select cells.
    """

    class Selected(Message):
        def __init__(self, table: "Table"):
            super().__init__()
            self.table = table

    def key_enter(self, event: events.Key) -> None:
        """
        Callback for pressing enter key.
        """
        return self.post_message(self.Selected(self))


class ExecutionList(DataTable):
    """
    Table of redun Executions.
    """

    executions = reactive([])

    class Selected(Message):
        def __init__(self, execution: Execution):
            super().__init__()
            self.execution = execution

    def __init__(self, executions: Optional[List[Execution]], **kwargs):
        super().__init__(**kwargs)
        self.cursor_type = "row"
        self.add_columns("ID", "Status", "Started", "Duration", "Execution")
        self.executions = executions

    def watch_executions(self) -> None:
        """
        Set the list of Executions to display.
        """
        self.clear()

        if self.executions is None:
            self.add_row("Loading...")
            return

        for execution in self.executions:
            self.add_row(
                f"Exec {execution.id[:8]}",
                style_status(execution.status_display),
                format_timestamp(execution.job.start_time),
                format_timedelta(execution.job.duration) if execution.job.duration else "",
                f"[[bold]{execution.job.task.namespace or 'no namespace'}[/bold]] "
                + " ".join(json.loads(execution.args)[1:]),
            )

    def key_enter(self, event: events.Key) -> None:
        """
        Callback for pressing enter key on an Execution.
        """
        if isinstance(self.executions, list):
            execution = self.executions[self.cursor_coordinate.row]
            return self.post_message(self.Selected(execution))


class JobList(DataTable):
    """
    Table of redun Jobs.
    """

    jobs = reactive([])

    class Selected(Message):
        def __init__(self, job: Job):
            super().__init__()
            self.job = job

    def __init__(self, jobs: Optional[List[Job]], **kwargs):
        super().__init__(**kwargs)
        self.cursor_type = "row"

        self.add_columns("ID", "Status", "Started", "Duration", "Job")
        self.jobs = jobs

    def watch_jobs(self) -> None:
        # Clear table.
        self.clear()

        if self.jobs is None:
            self.add_row("Loading...")
            return

        for job in self.jobs:
            self.add_row(
                f"{' ' * min(job.depth, 10)}Job {job.id[:8]}",
                style_status(job.status_display),
                format_timestamp(job.start_time),
                format_timedelta(job.duration) if job.duration else "",
                format_job(job),
            )

    def key_enter(self, event: events.Key) -> None:
        """
        Callback for pressing enter key on a Job.
        """
        job = self.jobs[self.cursor_coordinate.row]
        return self.post_message(self.Selected(job))


class JobStatusTable(Table):
    """
    Table of redun Job statuses by task.
    """

    JOB_STATUSES = [status for status in SchedulerJob.STATUSES if status != "PENDING"]

    def __init__(self, execution_id: str, **kwargs):
        super().__init__(**kwargs)
        self.execution_id = execution_id
        self.add_columns("TASK", *self.JOB_STATUSES)
        self.add_row("Loading...")

        self.tasks: List[Optional[Task]] = []

    def on_mount(self) -> None:
        self.call_after_refresh(self.load_jobs)

    async def load_jobs(self) -> None:
        # Clear loading message.
        self.clear()

        # Fetch jobs and task names from db.
        # Fetch Value.type in order to help compute Job status.
        self.job_tasks = (
            self.app.session.query(Job, Task.namespace, Task.name, Value.type)
            .outerjoin(CallNode, Job.call_hash == CallNode.call_hash)
            .outerjoin(Value, CallNode.value_hash == Value.value_hash)
            .join(Task, Job.task_hash == Task.hash)
            .filter(Job.execution_id == self.execution_id)
            .all()
        )

        tasks = (
            self.app.session.query(Task)
            .join(Job, Job.task_hash == Task.hash)
            .filter(Job.execution_id == self.execution_id)
            .all()
        )
        name2task = {task.fullname: task for task in tasks}

        # Compute status counts.
        job_status_counts: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(int))
        for job, namespace, name, result_type in self.job_tasks:
            fullname = namespace + "." + name if namespace else name
            job_status_counts[fullname][job.calc_status(result_type)] += 1
            job_status_counts[fullname]["TOTAL"] += 1
            job_status_counts["ALL"][job.status] += 1
            job_status_counts["ALL"]["TOTAL"] += 1

        # Populate table.
        task_names = sorted(job_status_counts.keys())
        for task_name in task_names:
            self.add_row(
                task_name,
                *[str(job_status_counts[task_name][status]) for status in self.JOB_STATUSES],
            )

        self.tasks = [name2task.get(task_name) for task_name in task_names]


class ValueSpan(Static):
    """
    Rendering of a redun Value that is linkable and collapsible.
    """

    expanded = reactive(False)

    def __init__(
        self,
        value: Value,
        format: str = "{value}",
        max_size: int = 100,
        on_click=lambda value_hash: None,
    ):
        super().__init__()
        self.value_hash = value.value_hash

        self.value_long = repr(value.preview)
        self.value_short = self.value_long[: max_size - len(format) + len("{value}")]

        self.format = format.format(value="[@click=click]{value}[/]")
        self.on_click = on_click

    def action_click(self) -> None:
        self.on_click(self.value_hash)

    def action_expand(self) -> None:
        self.expanded = not self.expanded
        self.refresh(layout=True)

    def render(self) -> Any:
        if self.value_short != self.value_long:
            if self.expanded:
                return self.format.format(value=self.value_long) + " [@click=expand](less)[/]"
            else:
                return self.format.format(value=self.value_short) + "[@click=expand]...[/]"
        else:
            return self.format.format(value=self.value_long)


class Interpreter(InteractiveInterpreter):
    """
    Interactive Python interpreter used by the ReplScreen.
    """

    def __init__(self, locals, write):
        super().__init__(locals)

        # Set output callback.
        self.write = write

    def run(self, code: str) -> Any:
        """
        Execute a code object.
        """
        try:
            try:
                # Try to evaluate as an expression first and return its result.
                return eval(code, cast(dict, self.locals))
            except SyntaxError:
                # If syntax error, it might be due to assignments being used.
                # Fallback to exec() instead.
                exec(code, cast(dict, self.locals))
                return NULL
        except SystemExit:
            raise
        except Exception:
            self.showtraceback()
            raise


class CommandInput(Input):
    """
    Input widget with history and tab-completion.
    """

    class Complete(Message):
        def __init__(self, text: str):
            super().__init__()
            self.text = text

    BINDINGS = [
        Binding("up", "cursor_up", "cursor up", show=False),
        Binding("down", "cursor_down", "cursor down", show=False),
    ]

    def __init__(self, *args, complete: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        if complete:
            self._bindings.bind("tab", "complete", "Complete", show=False)
        self.history: List[str] = []
        self.position = -1

    def on_input_submitted(self, message: Input.Submitted) -> None:
        # Add command to history.
        self.history.append(message.value)
        self.position = len(self.history)

    def action_cursor_up(self) -> None:
        if self.position > 0:
            # Move backwards through command history.
            self.position -= 1
            self.value = self.history[self.position]

    def action_cursor_down(self) -> None:
        if self.position < len(self.history):
            # Move forward in command history.
            self.position += 1
            if self.position < len(self.history):
                self.value = self.history[self.position]
            else:
                # When advancing off the end of history, give a blank command.
                self.value = ""

    def action_complete(self) -> None:
        """
        Callback for tab-complete attempt.
        """
        self.post_message(self.Complete(self.value))


class TagLinks(Container):
    """
    Displays links derived from redun Tags.
    """

    def __init__(self, link_patterns: List[str], tags: List[Tag]):
        super().__init__(classes="links")
        self.links = get_links(link_patterns, tags)

    def compose(self) -> ComposeResult:
        if self.links:
            yield Static("[bold]Links:[/]")
            for link in self.links:
                yield Static(f"  [#9999cc]{link}[/]")



import operator
from itertools import chain
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from redun.hashing import hash_arguments, hash_bytes, hash_struct
from redun.utils import iter_nested_value, pickle_dumps
from redun.value import Value, get_type_registry

# Type for an Expression or an arbitrary user value.
AnyExpression = Any
Result = TypeVar("Result")


def derive_expression(
    orig_expr: "Expression", derived_value: Union["Expression", Any]
) -> "Expression":
    """
    Record derived_value as downstream of orig_expression.
    """
    if not isinstance(derived_value, Expression):
        derived_expr: Expression = ValueExpression(derived_value)
    else:
        derived_expr = derived_value
    derived_expr._upstreams = [orig_expr]
    return derived_expr


class Expression(Value, Generic[Result]):
    """
    Base class for lazy expressions.

    Lazy expressions are used to defer compute until the redun scheduler is ready
    to execute it.
    """

    def __init__(self) -> None:
        self._hash: Optional[str] = None
        self._upstreams: List[Any] = []
        self._length: Optional[int] = None

    def get_hash(self, data: Optional[bytes] = None) -> str:
        if self._hash is None:
            self._hash = self._calc_hash()
        return self._hash

    def _calc_hash(self) -> str:
        raise NotImplementedError()

    def __getstate__(self) -> dict:
        return {}

    def __setstate__(self, state: dict) -> None:
        self._hash = None
        self._upstreams = []
        self._length = None

    def __call__(self, *args, **kwargs) -> "SimpleExpression":
        """need to keep structure for lazy function calls due to backwards compatibility"""
        return SimpleExpression("call", (self, args, kwargs))

    def __getitem__(self, key: Any) -> AnyExpression:
        """
        Provides lazy list and dict access by returning Expression.
        """
        if isinstance(key, slice):
            # Act like a list or tuple and return a slice of subpromises.
            subexprs = [self[i] for i in range(key.start or 0, key.stop, key.step or 1)]
            return subexprs

        else:
            return self._get_single_item(key)

    def __getattr__(self, key) -> "SimpleExpression":
        """needed for mypy; implemented via lazy operator"""
        raise NotImplementedError

    def _get_single_item(self, key) -> "SimpleExpression":
        """separate function needed for accessing single element; implemented via lazy operator"""
        raise NotImplementedError

    def __bool__(self) -> bool:
        raise TypeError("Expressions cannot be coerced to bool.")

    def __iter__(self) -> Iterator:
        if self._length:
            return (self[i] for i in range(self._length))
        raise TypeError("Expressions of unknown length cannot be iterated.")


class ApplyExpression(Expression[Result]):
    """
    Lazy expression for applying a function or task to arguments.
    """

    def __init__(self, args: Tuple, kwargs: dict):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        self._upstreams = [args, kwargs]


def format_arguments(args, kwargs) -> str:
    """
    Format Task arguments.
    """
    kwargs_items = sorted(kwargs.items())
    text = ", ".join(
        chain(
            (repr(arg) for arg in args),
            (f"{key}={repr(value)}" for key, value in kwargs_items),
        )
    )
    return text


class TaskExpression(ApplyExpression[Result]):
    """
    Lazy expression for applying a task to arguments.
    """

    def __init__(
        self,
        task_name: str,
        args: Tuple,
        kwargs: dict,
        task_options: Optional[dict] = None,
        length: Optional[int] = None,
    ):
        super().__init__(args, kwargs)
        self.task_name = task_name
        self.call_hash: Optional[str] = None
        self.task_expr_options = task_options or {}
        self._length = length

    def __repr__(self) -> str:
        return f"{self.task_name}({format_arguments(self.args, self.kwargs)})"

    def _calc_hash(self) -> str:
        registry = get_type_registry()
        args_hash = hash_arguments(registry, self.args, self.kwargs)
        task_options_hash = hash_bytes(pickle_dumps(self.task_expr_options))
        return hash_struct(["TaskExpression", self.task_name, args_hash, task_options_hash])

    def __getstate__(self) -> dict:
        state = super().__getstate__()
        registry = get_type_registry()
        return {
            **state,
            "task_name": self.task_name,
            "args": registry.serialize(self.args),
            "kwargs": registry.serialize(self.kwargs),
            "task_options": self.task_expr_options,
        }

    def __setstate__(self, state: dict) -> None:
        super().__setstate__(state)
        registry = get_type_registry()
        self.task_name = state["task_name"]
        self.args = registry.deserialize("builtins.tuple", state["args"])
        self.kwargs = registry.deserialize("builtins.dict", state["kwargs"])
        self.task_expr_options = state.get("task_options", {})
        self._upstreams = [self.args, self.kwargs]

    def is_valid(self) -> bool:
        from redun.task import get_task_registry

        return bool(get_task_registry().get(self.task_name)) and all(
            not isinstance(value, Value) or value.is_valid()
            for value in iter_nested_value((self.args, self.kwargs))
        )


class SimpleExpression(ApplyExpression[Result]):
    """
    Lazy expression for a simple computation (e.g. getattr, getitem, call).
    """

    def __init__(self, func_name: str, args: Tuple = (), kwargs: dict = {}):
        super().__init__(args, kwargs)
        self.func_name = func_name

    def __repr__(self) -> str:
        if self.func_name in _operator_name2symbol and len(self.args) == 2:
            # Binary operator.
            left, right = self.args
            if self.func_name in _reverse_operators:
                left, right = right, left
            return f"({repr(left)} {_operator_name2symbol[self.func_name]} {repr(right)})"

        elif self.func_name == "getitem" and len(self.args) == 2:
            left, right = self.args
            return f"{repr(left)}[{repr(right)}]"

        elif self.func_name == "getattr" and len(self.args) == 2:
            left, right = self.args
            return f"{repr(left)}.{right}"

        elif self.func_name == "call" and len(self.args) == 3:
            this, args, kwargs = self.args
            return f"{repr(this)}({format_arguments(args, kwargs)})"

        else:
            return "SimpleExpression('{func_name}', {args}, {kwargs})".format(
                func_name=self.func_name, args=repr(self.args), kwargs=repr(self.kwargs)
            )

    def _calc_hash(self) -> str:
        registry = get_type_registry()
        args_hash = hash_arguments(registry, self.args, self.kwargs)
        return hash_struct(["SimpleExpression", self.func_name, args_hash])

    def __getstate__(self) -> dict:
        state = super().__getstate__()
        registry = get_type_registry()
        return {
            **state,
            "func_name": self.func_name,
            "args": registry.serialize(self.args),
            "kwargs": registry.serialize(self.kwargs),
        }

    def __setstate__(self, state: dict) -> None:
        super().__setstate__(state)
        registry = get_type_registry()
        self.func_name = state["func_name"]
        self.args = registry.deserialize("builtins.tuple", state["args"])
        self.kwargs = registry.deserialize("builtins.dict", state["kwargs"])
        self._upstreams = [self.args, self.kwargs]


class SchedulerExpression(TaskExpression[Result]):
    """
    Lazy expression that is evalulated within the scheduler for redun-specific operations.
    """

    def _calc_hash(self) -> str:
        registry = get_type_registry()
        args_hash = hash_arguments(registry, self.args, self.kwargs)
        return hash_struct(["SchedulerExpression", self.task_name, args_hash])


class ValueExpression(Expression[Result]):
    """
    Lifts a concrete value into an Expression type.
    """

    def __init__(self, value: Result):
        super().__init__()
        assert not isinstance(value, Expression)
        self.value = value

    def __repr__(self) -> str:
        return f"ValueExpression({repr(self.value)})"

    def _calc_hash(self) -> str:
        registry = get_type_registry()
        value_hash = registry.get_hash(self.value)
        return hash_struct(["ValueExpression", value_hash])

    def __getstate__(self) -> dict:
        state = super().__getstate__()
        registry = get_type_registry()
        return {
            **state,
            "value_type": registry.get_type_name(type(self.value)),
            "value": registry.serialize(self.value),
        }

    def __setstate__(self, state: dict) -> None:
        super().__setstate__(state)
        registry = get_type_registry()
        self.value = registry.deserialize(state["value_type"], state["value"])


class QuotedExpression(Generic[Result]):
    """
    A quoted expression that does will not evaluate until forced.
    """

    def __init__(self, expr: Result):
        self._expr = expr

    def __repr__(self) -> str:
        return f"QuotedExpression({repr(self._expr)})"

    def eval(self) -> Result:
        """
        Evaluate the quoted expression.
        """
        return self._expr


def quote(expr: Result) -> QuotedExpression[Result]:
    """
    Quote an Expression so that it does not evaluate.
    """
    return QuotedExpression(expr)


_lazy_operation_registry: Dict[str, Callable] = {}
_operator_name2symbol = {}
_reverse_operators = set()


def get_lazy_operation(name: str) -> Optional[Callable]:
    """
    Retrieves lazy operation by registered name
    """
    return _lazy_operation_registry.get(name)


def lazy_operation(
    method: Optional[str] = None,
    name: Optional[str] = None,
    symbol: Optional[str] = None,
    reverse: bool = False,
) -> Callable[[Callable], Callable]:
    """
    Function decorator to declare lazy operations on Expression object.

    Args:
        method: Name of method registered to Expression class. Defaults to None.
        name: Registered name of operation by which it will be retrieved. Defaults to None.
    """

    def deco(func: Callable) -> Callable:
        _method: str = method or func.__name__
        _name: str = name or func.__name__

        def wrapper(self, *args, **kwargs) -> SimpleExpression:
            return SimpleExpression(_name, (self, *args), kwargs)

        setattr(Expression, _method, wrapper)
        _lazy_operation_registry[_name] = func
        if symbol:
            _operator_name2symbol[_name] = symbol
        if reverse:
            _reverse_operators.add(_name)

        return func

    return deco


# Register standard operators as lazy operations on Expressions.
lazy_operation(method="__eq__", symbol="==")(operator.eq)
lazy_operation(method="__ne__", symbol="!=")(operator.ne)
lazy_operation(method="__lt__", symbol="<")(operator.lt)
lazy_operation(method="__le__", symbol="<=")(operator.le)
lazy_operation(method="__gt__", symbol=">")(operator.gt)
lazy_operation(method="__ge__", symbol=">=")(operator.ge)
lazy_operation(method="__add__", symbol="+")(operator.add)
lazy_operation(method="__radd__", name="radd", symbol="+", reverse=True)(
    lambda self, other: other + self
)
lazy_operation(method="__sub__", symbol="-")(operator.sub)
lazy_operation(method="__rsub__", name="rsub", symbol="-", reverse=True)(
    lambda self, other: other - self
)
lazy_operation(method="__mul__", symbol="*")(operator.mul)
lazy_operation(method="__rmul__", name="rmul", symbol="*", reverse=True)(
    lambda self, other: other * self
)
lazy_operation(method="__truediv__", name="div", symbol="/")(operator.truediv)
lazy_operation(method="__rtruediv__", name="rdiv", symbol="/", reverse=True)(
    lambda self, other: other / self
)
lazy_operation(method="__and__", name="and", symbol="&")(lambda self, other: self and other)
lazy_operation(method="__rand__", name="rand", symbol="&", reverse=True)(
    lambda self, other: other and self
)
lazy_operation(method="__or__", name="or", symbol="|")(lambda self, other: self or other)
lazy_operation(method="__ror__", name="ror", symbol="|", reverse=True)(
    lambda self, other: other or self
)


def _lazy_call(self, args=(), kwargs={}):
    """
    Lazy evaluation for callable.
    """
    return self(*args, **kwargs)


_lazy_operation_registry["call"] = _lazy_call


@lazy_operation(name="getattr", method="__getattr__")
def _lazy_getattr(self, field: str):
    """
    Lazy evaluation for namedtuple.
    """
    return getattr(self, field)


@lazy_operation(name="getitem", method="_get_single_item")
def _lazy_getitem(self, key: Any):
    """
    Provides lazy list and dict access.
    """
    return self[key]



from typing import Any, Callable

from redun.task import Task, wraps_task


def VarArg(type=Any):
    """A *args-style variadic positional argument"""
    return type


def KwArg(type=Any):
    """A **kwargs-style variadic keyword argument"""
    return type


def square_task():
    """A trivial wrapper that is not defined in the same module as the test."""

    @wraps_task(wrapper_name="square_task")
    def _square_task(inner_task: Task) -> Callable[[VarArg(Any), KwArg(Any)], Any]:
        def do_square(*args, **kwargs):
            return inner_task.func(*args, **kwargs) ** 2

        return do_square

    return _square_task



"""add job and value indexes

Revision ID: f68b3aaee9cc
Revises: eb7b95e4e8bf
Create Date: 2024-06-14 18:42:22.604907

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "f68b3aaee9cc"
down_revision = "eb7b95e4e8bf"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("job", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_job_end_time"), ["end_time"], unique=False)
        batch_op.create_index(batch_op.f("ix_job_start_time"), ["start_time"], unique=False)

    with op.batch_alter_table("value", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_value_type"), ["type"], unique=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("value", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_value_type"))

    with op.batch_alter_table("job", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_job_start_time"))
        batch_op.drop_index(batch_op.f("ix_job_end_time"))

    # ### end Alembic commands ###



import os
import threading
import time
from typing import List, cast

import pytest

from redun import File, Scheduler, task
from redun.config import Config
from redun.executors.local import LocalExecutor
from redun.tests.utils import use_tempdir

redun_namespace = "redun.tests.test_executors"


@task()
def task1(x: int, wait: int, count_file: File) -> int:
    with count_file.open("a") as out:
        out.write("x")
    while len(count_file.read()) < wait:
        time.sleep(0.01)
    return os.getpid()


@task()
def main(count_file: File) -> List[int]:
    return [task1(1, 2, count_file), task1(2, 2, count_file)]


@task()
def bad_task():
    raise ValueError("boom")


@use_tempdir
def test_process_executor() -> None:
    """
    Ensure tasks can run in different processes.
    """
    config = Config({"executors.default": {"type": "local", "mode": "process"}})
    scheduler = Scheduler(config=config)

    assert cast(LocalExecutor, scheduler.executors["default"]).mode == "process"

    # Use the length of file to count that all processes have run before exiting.
    # This ensures we get unique pids.
    count_file = File("count_file")
    count_file.touch()
    [pid1, pid2] = scheduler.run(main(count_file))
    assert pid1 != pid2

    # Errors should be raised across processes.
    with pytest.raises(ValueError):
        scheduler.run(bad_task())


@task(executor="process")
def task2(x: int, wait: int, count_file: File) -> int:
    with count_file.open("a") as out:
        out.write("x")
    while len(count_file.read()) < wait:
        time.sleep(0.01)
    return os.getpid()


@task()
def main2(count_file: File) -> List[int]:
    return [task2(1, 2, count_file), task2(2, 2, count_file)]


@use_tempdir
def test_process_executor_options() -> None:
    """
    Choose local executor mode using task_options.
    """
    scheduler = Scheduler()
    assert cast(LocalExecutor, scheduler.executors["default"]).mode == "thread"

    # Use the length of file to count that all processes have run before exiting.
    # This ensures we get unique pids.
    count_file = File("count_file")
    count_file.touch()
    [pid1, pid2] = scheduler.run(main2(count_file))
    assert pid1 != pid2


@task(executor="process", script=True)
def task3(x: int, wait: int, count_file: File) -> str:
    with count_file.open("a") as out:
        out.write("x")
    while len(count_file.read()) < wait:
        time.sleep(0.01)
    pid = os.getpid()
    return "echo {}".format(pid)


@task()
def main3(count_file: File) -> List[str]:
    return [task3(1, 2, count_file), task3(2, 2, count_file)]


@use_tempdir
def test_process_executor_script() -> None:
    """
    Ensure script tasks can run in different processes.
    """

    config = Config({"executors.default": {"type": "local", "mode": "process"}})
    scheduler = Scheduler(config=config)

    assert cast(LocalExecutor, scheduler.executors["default"]).mode == "process"

    # Use the length of file to count that all processes have run before exiting.
    # This ensures we get unique pids.
    count_file = File("count_file")
    count_file.touch()
    [pid1, pid2] = scheduler.run(main3(count_file))
    assert pid1 != pid2
    assert pid1.strip().isdigit()
    assert pid2.strip().isdigit()


@task(executor="process")
def process_task(x: int):
    time.sleep(0.1)
    return {"thread_count": threading.active_count(), "pid": os.getpid()}


@task()
def thread_task(x: int) -> dict:
    return {
        "thread_count": threading.active_count(),
        "process": process_task(x),
        "data": x,
    }


@task()
def main4() -> List[dict]:
    return [thread_task(i) for i in range(10)]


def test_thread_count() -> None:
    """
    Ensure there are no extra threads in subprocesses.

    If creating a new process copies all the extra threads it could lead to
    unstable results.
    """
    scheduler = Scheduler()
    results = scheduler.run(main4())

    for row in results:
        assert row["process"]["thread_count"] == 1

    # The thread tasks will be in a process with multiple threads.
    assert max(row["thread_count"] for row in results) > 1



import time
from typing import List

import pytest

from redun import Scheduler, task
from redun.backends.db import Execution, Job, Session
from redun.functools import (
    apply_func,
    as_task,
    compose,
    const,
    delay,
    eval_,
    flat_map,
    flatten,
    force,
    identity,
    map_,
    seq,
    starmap,
    zip_,
)


def test_compose(scheduler: Scheduler) -> None:
    """
    compose() should be able to compose multiple tasks together.
    """

    @task()
    def task1(x: int) -> int:
        return x + 1

    @task()
    def task2(x: int) -> int:
        return x * 2

    task3 = compose(task2, task1)
    assert scheduler.run(task3(1)) == 4

    task4 = compose(task1, task2, task1)
    assert scheduler.run(task4(1)) == 5


def test_delay(scheduler: Scheduler) -> None:
    """
    delay() should delay evaluation of its argument. force() should force evaluation.
    """
    calls = []

    @task()
    def task1():
        calls.append("task1")
        return 10

    @task()
    def main(x):
        return x

    expr = delay(task1())
    scheduler.run(main(expr))
    assert calls == []

    assert scheduler.run(main(force(expr))) == 10
    assert calls == ["task1"]


def test_seq(scheduler: Scheduler) -> None:
    """
    seq() should sequence the execution of tasks.
    """
    calls = []

    @task()
    def subtask(name, interval):
        time.sleep(interval)
        calls.append(name)
        return name

    assert scheduler.run(
        seq(
            [
                subtask(1, 0.2),
                subtask(2, 0.1),
                subtask(3, 0.01),
                subtask(4, 0.01),
            ]
        )
    ) == [1, 2, 3, 4]
    assert calls == [1, 2, 3, 4]

    assert scheduler.run(seq([])) == []
    assert scheduler.run(seq([10])) == [10]
    assert scheduler.run(seq([subtask(1, 0)])) == [1]


def test_const(scheduler: Scheduler) -> None:
    """
    const() should discard the second argument.
    """

    assert scheduler.run(const(1, 2)) == 1
    assert scheduler.run(const(2, 1)) == 2


def add(a: int, b: int) -> int:
    return a + b


@task(namespace="redun.tests")
def range_(start, stop):
    return list(range(start, stop))


def test_apply_func(scheduler: Scheduler) -> None:
    """
    apply_func() should call a plain python function on lazy args.
    """
    assert scheduler.run(apply_func(max, 30, 20, 10)) == 30
    assert scheduler.run(apply_func(add, 10, b=20)) == 30
    assert scheduler.run(apply_func(add, a=2, b=20)) == 22

    assert scheduler.run(apply_func(max, range_(1, 4))) == 3


def test_as_task(scheduler: Scheduler) -> None:
    """
    as_task() should convert a python function to a redun Task.
    """
    assert scheduler.run(as_task(max)(range_(1, 4))) == 3

    inc = as_task(add).partial(1)
    assert scheduler.run(map_(inc, range_(1, 4))) == [2, 3, 4]
    assert scheduler.run(as_task(len)(range_(1, 4))) == 3


def test_eval(scheduler: Scheduler) -> None:
    """
    eval_() should evaluate a string of python.
    """

    # Evaluate a python string.
    assert scheduler.run(eval_("1 + 2")) == 3

    # Use keyword arguments.
    assert scheduler.run(eval_("{a, b}", a=1, b=2)) == {1, 2}

    # Use a positional argument.
    assert scheduler.run(eval_("x + 2", 1, pos_args=["x"])) == 3

    # Use a positional argument.
    assert scheduler.run(eval_("x + y", 1, 2, pos_args=["x", "y"])) == 3

    # Disallow too many positional arguments.
    with pytest.raises(TypeError):
        scheduler.run(eval_("x + y", 1, 2, pos_args=["x"]))

    # Allow indented code.
    assert scheduler.run(
        eval_(
            """
                {a, b}
                """,
            a=1,
            b=2,
        )
    ) == {1, 2}

    # eval_ should work with partial application.
    assert scheduler.run(eval_.partial("{a, b}")(a=1, b=2)) == {1, 2}
    assert scheduler.run(eval_.partial("x + 1", pos_args=["x"])(5)) == 6


def test_map(scheduler: Scheduler, session: Session) -> None:
    """
    map_() should perform function composition optimizations.
    """

    @task()
    def inc(x: int) -> int:
        return x + 1

    @task()
    def double(x: int) -> int:
        return 2 * x

    @task()
    def get_values() -> List[int]:
        return [1, 2, 3]

    # Simple map over a list.
    assert scheduler.run(map_(inc, [1, 2, 3])) == [2, 3, 4]

    # List of values can be an expression as well.
    assert scheduler.run(map_(inc, get_values())) == [2, 3, 4]

    # Task can be an expression.
    assert scheduler.run(map_(identity(inc), [1, 2, 3])) == [2, 3, 4]

    # Multiple maps should be composed.
    @task()
    def main() -> List[int]:
        return map_(inc, map_(double, [1, 2, 3]))

    assert scheduler.run(main()) == [3, 5, 7]

    exec1 = (
        session.query(Execution)
        .join(Job, Job.id == Execution.job_id)
        .order_by(Job.start_time.desc())
        .first()
    )

    for job in exec1.job.child_jobs:
        assert job.task.name == "compose_apply"
        assert [child_job.task.name for child_job in job.child_jobs] == [
            "double",
            "inc",
        ]


def test_starmap(scheduler: Scheduler, session: Session) -> None:
    """
    starmap_() should map a task over a list of keyword arguments.
    """

    kwargs = [
        {"a": 1, "b": 2},
        {"a": 3, "b": 4},
    ]

    @task()
    def add(a, b):
        return a + b

    assert scheduler.run(starmap(add, kwargs)) == [3, 7]

    @task()
    def get_kwargs():
        return [
            {"a": 1, "b": 2},
            {"a": 3, "b": 4},
        ]

    # starmap should be able to take kwargs as an Expression.
    assert scheduler.run(starmap(add, get_kwargs())) == [3, 7]

    @task()
    def swap(a, b):
        return {"a": b, "b": a}

    # Multiple maps should be composed.
    @task()
    def main() -> List[int]:
        return starmap(add, starmap(swap, kwargs))

    assert scheduler.run(main()) == [3, 7]

    exec1 = (
        session.query(Execution)
        .join(Job, Job.id == Execution.job_id)
        .order_by(Job.start_time.desc())
        .first()
    )

    for job in exec1.job.child_jobs:
        assert job.task.name == "compose_apply"
        assert [child_job.task.name for child_job in job.child_jobs] == [
            "eval_",
            "eval_",
        ]
        assert [child_job.child_jobs[0].task.name for child_job in job.child_jobs] == [
            "swap",
            "add",
        ]


def test_flatten(scheduler: Scheduler) -> None:
    """
    flatten() should flatten lists.
    """

    @task()
    def split(text: str) -> List[str]:
        return text.split()

    texts = ["hello world", "not invented here"]
    assert scheduler.run(flatten(map_(split, texts))) == [
        "hello",
        "world",
        "not",
        "invented",
        "here",
    ]
    assert scheduler.run(flat_map(split, texts)) == [
        "hello",
        "world",
        "not",
        "invented",
        "here",
    ]


def test_zip(scheduler: Scheduler) -> None:
    """
    zip_() should zip lists.
    """

    @task()
    def range_(a: int, b: int) -> List[int]:
        return list(range(a, b))

    assert scheduler.run(zip_(range_(0, 5), range_(1, 6))) == [
        (0, 1),
        (1, 2),
        (2, 3),
        (3, 4),
        (4, 5),
    ]



import os
from unittest.mock import MagicMock, patch

import pytest

from redun import azure_utils


@patch("redun.azure_utils.AzureMLOnBehalfOfCredential")
@patch("redun.azure_utils.DefaultAzureCredential")
@patch("redun.azure_utils.is_azure_managed_node", return_value=False)
def test_azure_default_credential_outside_azure(
    _, default_credential_mock, on_behalf_credential_mock
):
    cr = azure_utils.get_az_credential()
    default_credential_mock.assert_called_with(exclude_managed_identity_credential=True)
    assert cr == default_credential_mock.return_value
    on_behalf_credential_mock.assert_not_called()


@patch("redun.azure_utils.AzureMLOnBehalfOfCredential")
@patch("redun.azure_utils.DefaultAzureCredential")
@patch("redun.azure_utils.is_azure_managed_node", return_value=True)
def test_azure_aml_credential_inside_azure(_, default_credential_mock, on_behalf_credential_mock):
    cr = azure_utils.get_az_credential()
    assert cr == on_behalf_credential_mock.return_value
    default_credential_mock.assert_not_called()


@patch("redun.azure_utils.AzureMLOnBehalfOfCredential")
@patch("redun.azure_utils.DefaultAzureCredential")
@patch("redun.azure_utils.is_azure_managed_node", return_value=True)
def test_azure_default_credential_inside_azure(
    _, default_credential_mock, on_behalf_credential_mock
):
    on_behalf_credential_mock.return_value.get_token = MagicMock(
        side_effect=Exception("failed to get token")
    )

    cr = azure_utils.get_az_credential()
    assert cr == default_credential_mock.return_value
    on_behalf_credential_mock.assert_called_once()


@patch("redun.azure_utils.AzureMLOnBehalfOfCredential")
@patch("redun.azure_utils.DefaultAzureCredential")
@patch("redun.azure_utils.is_azure_managed_node", return_value=True)
def test_azure_default_identity_fallback_credential(
    _, default_credential_mock, on_behalf_credential_mock
):
    os.environ["DEFAULT_IDENTITY_CLIENT_ID"] = "123-456-789"
    on_behalf_credential_mock.return_value.get_token = MagicMock(
        side_effect=Exception("failed to get token")
    )
    cr = azure_utils.get_az_credential()
    del os.environ["DEFAULT_IDENTITY_CLIENT_ID"]
    assert cr == default_credential_mock.return_value
    default_credential_mock.assert_called_once_with(managed_identity_client_id="123-456-789")

    on_behalf_credential_mock.assert_called_once()


@patch("redun.azure_utils.AzureMLOnBehalfOfCredential")
@patch("redun.azure_utils.DefaultAzureCredential")
@patch("redun.azure_utils.is_azure_managed_node", return_value=True)
def test_azure_aml_job_identity_used_before_default_identity(
    _, default_credential_mock, on_behalf_credential_mock
):
    os.environ["AZUREML_USER_MANAGED_CLIENT_ID"] = "0000-0000"
    os.environ["DEFAULT_IDENTITY_CLIENT_ID"] = "123-456-789"
    on_behalf_credential_mock.return_value.get_token = MagicMock(
        side_effect=Exception("failed to get token")
    )
    cr = azure_utils.get_az_credential()
    del os.environ["DEFAULT_IDENTITY_CLIENT_ID"]
    del os.environ["AZUREML_USER_MANAGED_CLIENT_ID"]
    assert cr == default_credential_mock.return_value
    default_credential_mock.assert_called_once_with(managed_identity_client_id="0000-0000")

    on_behalf_credential_mock.assert_called_once()


@patch("redun.azure_utils.AzureMLOnBehalfOfCredential")
@patch("redun.azure_utils.DefaultAzureCredential")
@patch("redun.azure_utils.is_azure_managed_node", return_value=True)
def test_azure_client_id_variable_takes_precedence(
    _, default_credential_mock, on_behalf_credential_mock
):
    os.environ["AZURE_CLIENT_ID"] = "111-111"
    os.environ["AZUREML_USER_MANAGED_CLIENT_ID"] = "0000-0000"
    os.environ["DEFAULT_IDENTITY_CLIENT_ID"] = "123-456-789"
    on_behalf_credential_mock.return_value.get_token = MagicMock(
        side_effect=Exception("failed to get token")
    )
    cr = azure_utils.get_az_credential()
    del os.environ["DEFAULT_IDENTITY_CLIENT_ID"]
    del os.environ["AZUREML_USER_MANAGED_CLIENT_ID"]
    del os.environ["AZURE_CLIENT_ID"]
    assert cr == default_credential_mock.return_value
    default_credential_mock.assert_called_once_with(managed_identity_client_id="111-111")

    on_behalf_credential_mock.assert_called_once()


@patch("redun.azure_utils.AzureMLOnBehalfOfCredential")
@patch("redun.azure_utils.DefaultAzureCredential")
@patch("redun.azure_utils.is_azure_managed_node", return_value=True)
def test_azure_raise_error_if_all_credentials_fail_to_get_token(
    _, default_credential_mock, on_behalf_credential_mock
):
    """
    This test checks if in case of error getting AML credential
    and no default managed identity, the error is properly raised.
    """
    default_credential_mock.return_value.get_token = MagicMock(
        side_effect=Exception("failed to get token")
    )
    on_behalf_credential_mock.return_value.get_token = MagicMock(
        side_effect=Exception("failed to get token")
    )

    with pytest.raises(Exception):
        _ = azure_utils.get_az_credential()

    default_credential_mock.assert_called_once()
    on_behalf_credential_mock.assert_called_once()



"""add updated_time field

Revision ID: 0bee3d6dba76
Revises: 3b0a6e67cc58
Create Date: 2024-10-01 12:21:00.709029

"""

from alembic import op
import sqlalchemy as sa

from redun.backends.db import DateTimeUTC

# revision identifiers, used by Alembic.
revision = "0bee3d6dba76"
down_revision = "3b0a6e67cc58"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("execution", schema=None) as batch_op:
        batch_op.add_column(sa.Column("updated_time", DateTimeUTC(timezone=True), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("execution", schema=None) as batch_op:
        batch_op.drop_column("updated_time")

    # ### end Alembic commands ###



"""Sphinx documentation plugin used to document redun tasks.

Introduction
============

Usage
-----

Add the extension to your :file:`docs/conf.py` configuration module:

.. code-block:: python

    extensions = (...,
                  'redun.sphinx')

If you'd like to change the prefix for tasks in reference documentation
then you can change the ``redun_task_prefix`` configuration value:

.. code-block:: python

    redun_task_prefix = '(task)'  # < default

With the extension installed `autodoc` will automatically find
task decorated objects (e.g. when using the automodule directive)
and generate the correct (as well as add a ``(task)`` prefix),
and you can also refer to the tasks using `:task:proj.tasks.add`
syntax.

Use ``.. autotask::`` to alternatively manually document a task.
"""
# This technique is inspired by:
# https://docs.celeryq.dev/en/latest/_modules/celery/contrib/sphinx.html

from inspect import getfullargspec, signature
from typing import Any, List, Optional

from docutils import nodes
from sphinx.domains.python import PyFunction
from sphinx.ext.autodoc import FunctionDocumenter
from sphinx.util.docstrings import prepare_docstring
from sphinx.util.inspect import getdoc

from redun.task import SchedulerTask, Task


class TaskDocumenter(FunctionDocumenter):
    """Document redun task definitions."""

    objtype = "task"
    member_order = 11

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent) -> bool:
        # This Documenter documents objects of type Task.
        return isinstance(member, Task) and not isinstance(member, SchedulerTask)

    def format_args(self, **kwargs: Any) -> str:
        """
        Returns a string documenting the task's parameters.
        """
        wrapped = self.object.func
        if wrapped is not None:
            return str(signature(wrapped))
        return ""

    def get_doc(self, ignore: Optional[int] = None) -> List[List[str]]:
        """
        Returns the formatted docstring for the task.
        """
        docstring = getdoc(
            self.object.func,
            self.get_attr,
            self.config.autodoc_inherit_docstrings,
            self.parent,
            self.object_name,
        )
        if docstring:
            tab_width = self.directive.state.document.settings.tab_width
            return [prepare_docstring(docstring, tab_width)]
        return []

    def document_members(self, all_members=False):
        pass

    def check_module(self) -> bool:
        # Normally checks if *self.object* is really defined in the module
        # given by *self.modname*. But since functions decorated with the @task
        # decorator are instances living in the redun.task, we have to check
        # the wrapped function instead.
        if self.object.func.__module__ == self.modname:
            return True
        return super().check_module()


class TaskDirective(PyFunction):
    """Sphinx task directive."""

    def get_signature_prefix(self, sig: str) -> List[nodes.Node]:
        return [nodes.Text(self.env.config.redun_task_prefix)]


class SchedulerTaskDocumenter(TaskDocumenter):
    """Document redun scheduler_task definitions."""

    objtype = "scheduler_task"
    member_order = 11

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent) -> bool:
        # This Documenter documents objects of type Task.
        return isinstance(member, SchedulerTask)

    def format_args(self, **kwargs: Any) -> str:
        """
        Returns a string documenting the task's parameters.
        """
        wrapped = self.object.func
        if wrapped is not None:
            sig = signature(wrapped)

            # Remove SchedulerTask-specific internal parameters.
            params = list(sig.parameters.values())[3:]

            # Unwrap the Promise return value.
            argspec = getfullargspec(wrapped)
            return_annotation = argspec.annotations["return"].__parameters__[0]

            return str(sig.replace(parameters=params, return_annotation=return_annotation))
        return ""


class SchedulerTaskDirective(PyFunction):
    """Sphinx scheduler_task directive."""

    def get_signature_prefix(self, sig: str) -> List[nodes.Node]:
        return [nodes.Text(self.env.config.redun_scheduler_task_prefix)]


def autodoc_skip_member_handler(app, what, name, obj, skip, options) -> Optional[bool]:
    """Handler for autodoc-skip-member event."""
    # redun tasks created with the @task decorator have the property
    # that *obj.__doc__* and *obj.__class__.__doc__* are equal, which
    # trips up the logic in sphinx.ext.autodoc that is supposed to
    # suppress repetition of class documentation in an instance of the
    # class. This overrides that behavior.
    if isinstance(obj, Task):
        if skip:
            return False
    return None


def setup(app):
    """Setup Sphinx extension."""
    app.setup_extension("sphinx.ext.autodoc")

    app.add_autodocumenter(TaskDocumenter)
    app.add_directive_to_domain("py", "task", TaskDirective)

    app.add_autodocumenter(SchedulerTaskDocumenter)
    app.add_directive_to_domain("py", "scheduler_task", SchedulerTaskDirective)

    app.add_config_value("redun_task_prefix", "(task)", True)
    app.add_config_value("redun_scheduler_task_prefix", "(scheduler_task)", True)
    app.connect("autodoc-skip-member", autodoc_skip_member_handler)

    return {"parallel_read_safe": True}
